﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Dandelion
{
	/// <summary>
	/// Collection of common functions/definitions.
	/// </summary>
	public class AODCommon
	{
		#region Status Code Definitions

		/// <summary>
		/// Specifies the status code of dandelion response.
		/// </summary>
		public enum DandelionStatusCode : int
		{
			/// <summary>
			/// The server encountered an unexpected condition which prevented it from fulfilling the request.
			/// </summary>
			ControllerStatusCodeServerFailure = 1000,

			/// <summary>
			/// An WebException occurred except for the WebExceptionStatus.ReceiveFailure.
			/// </summary>
			ControllerStatusCodeInternetFailure = 1001,

			/// <summary>
			/// An Exception occurred except for the wifi.
			/// </summary>
			ControllerStatusCodeWifiFailure = 1002,

			/// <summary>
			/// An Exception occurred except for timeout.
			/// </summary>
			ControllerStatusCodeTimeout = 1003
		};

		#endregion


		#region Enum/Handler/Event for StartPairing

		/// <summary>
		/// Handler for paring.
		/// </summary>
		public delegate void PairingEventHandler(object sender, PairingResultEventArgs e);

		/// <summary>
		/// Represents the pairing start event data.
		/// </summary>
		public class PairingResultEventArgs : EventArgs
		{
			/// <summary>
			/// Gets a value indicating whether this operation is success.
			/// </summary>
			public bool IsSuccess { get; internal set; }

			/// <summary>
			/// Gets the instance id.
			/// </summary>
			public string InstanceID { get; internal set; }

			/// <summary>
			/// Gets the profile.
			/// </summary>
			public string Profile { get; internal set; }

			/// <summary>
			/// Gets the bool value for check the device is final or not.
			/// </summary>
			public bool IsCompleted { get; internal set; }

			/// <summary>
			/// Gets the status code.
			/// </summary>
			public DandelionStatusCode StatusCode { get; internal set; }

			/// <summary>
			/// Gets the exception.
			/// </summary>
			public Exception Exception { get; internal set; }
		}

		#endregion


		#region Enum/Handler/Event for Login

		/// <summary>
		/// Specifies the type of login.
		/// </summary>
		public enum LoginType : int
		{
			None = 0,
			Goolge = 1,
			Facebook = 2,
			Acer = 3
		};

		/// <summary>
		/// Represents the result of login status.
		/// </summary>
		public class LoginStatusResult
		{
			public string Username { get; internal set; }
			public LoginType Type { get; internal set; }
		}

		public delegate void LoginResultEventHandler(object sender, LoginResultEventArgs args);
		public delegate void LogoutResultEventHandler(object sender, LogoutResultEventArgs args);

		public class LoginResultEventArgs : EventArgs
		{
			public LoginType Type { get; internal set; }
			public string Username { get; internal set; }
			public bool IsSuccess { get; internal set; }
			public Exception Exception { get; internal set; }
		}

		public class LogoutResultEventArgs : EventArgs
		{
			public bool IsSuccess { get; internal set; }
			public Exception Exception { get; internal set; }
		}

		#endregion


		#region Enum/Handler/Event for Notification

		public delegate void NotificationResultEventHandler(object sender, NotificationResultEventArgs args);

		public class NotificationResultEventArgs : EventArgs
		{
			/// <summary>
			/// Gets a value indicating whether this operation is success.
			/// </summary>
			public bool IsSuccess { get; internal set; }

			/// <summary>
			/// Gets the information.
			/// </summary>
			public string Info { get; internal set; }

			/// <summary>
			/// Gets the body.
			/// </summary>
			public string Body { get; internal set; }

			/// <summary>
			/// Gets the exception.
			/// </summary>
			public Exception Exception { get; internal set; }
		}

		#endregion


		#region TaskResult Definitions for Controller API

		/// <summary>
		/// Represents the task result arguments of general.
		/// </summary>
		public abstract class TaskResult
		{
			/// <summary>
			/// Gets a value indicating whether this operation is success.
			/// </summary>
			public bool IsSuccess { get; internal set; }

			/// <summary>
			/// Gets the status code.
			/// </summary>
			public DandelionStatusCode StatusCode { get; internal set; }

			/// <summary>
			/// Gets the exception.
			/// </summary>
			public Exception Exception { get; internal set; }
		}

		/// <summary>
		/// Represents the task result arguments of profile.
		/// </summary>
		public abstract class ProfileTaskResult
		{
			/// <summary>
			/// Gets the profile.
			/// </summary>
			public string Profile { get; internal set; }

			/// <summary>
			/// Gets a value indicating whether this operation is success.
			/// </summary>
			public bool IsSuccess { get; internal set; }

			/// <summary>
			/// Gets the status code.
			/// </summary>
			public DandelionStatusCode StatusCode { get; internal set; }

			/// <summary>
			/// Gets the exception.
			/// </summary>
			public Exception Exception { get; internal set; }
		}

		/// <summary>
		/// Represents the task result arguments of initialize device.
		/// </summary>
		public class InitializeTaskResult : TaskResult { }

		/// <summary>
		/// Represents the task result arguments of update device profile.
		/// </summary>
		public class UpdateDeviceProfileResult : ProfileTaskResult { }

		/// <summary>
		/// Represents the task result arguments of get device profile.
		/// </summary>
		public class GetDeviceProfileResult : ProfileTaskResult { }

		/// <summary>
		/// Represents the task result arguments of pairing request.
		/// </summary>
		public class PairingInvitationTaskResult : TaskResult { }

		/// <summary>
		/// Represents the task result arguments of set remote message action.
		/// </summary>
		public class SendMessageTaskResult : TaskResult { }

		#endregion
	}
}
