﻿using Microsoft.Win32;
using System;

namespace AirFusion.ConnectionManager
{
	/// <summary>
	/// Check custom Registry setting value about remote device has Permission or not
	/// </summary>
	public enum RemoteDevicePermission
	{
		AutoDetect,
		HavePermission,
		NoPermission,
		NotDetect,
	}

	public enum VideoProtocol
	{
		UDP = 1,
		TCP = 2,
	}

	public static class CustomRegSetting
	{
		/// <summary>
		/// Gets the remote device permission.
		/// </summary>
		/// <returns>
		/// If return "RemoteDevicePermission.AutoDetect" represent we will based on Device Info to determin the device has/not System Permission
		/// If return "RemoteDevicePermission.HavePermission" (for debug use), represent the device has System Permission
		/// If return "RemoteDevicePermission.NoPermission" (for debug use), represent the device hadn't System Permission
		/// </returns>
		public static RemoteDevicePermission GetRemoteDevicePermission()
		{
			// Get phone system permission detection rule from registry
			try
			{
				bool isX86;
				object oRK = RegistryUtility.GetRegistry(RegistryHive.CurrentUser, Configs.RegVideoPath, Configs.RegPhoneSystemPermissionDetectionRuleKey, RegistryView.Registry64, out isX86);

				if (oRK != null)
				{
					return (RemoteDevicePermission)oRK;
				}
			}
			catch
			{
				return RemoteDevicePermission.AutoDetect;
			}

			return RemoteDevicePermission.AutoDetect;
		}

		/// <summary>
		/// Gets the size of the video.
		/// </summary>
		/// <param name="videoWidth">Width of the video.</param>
		/// <param name="videoHeight">Height of the video.</param>
		public static void GetVideoSize(ref int videoWidth, ref int videoHeight)
		{
			bool isX86;
			// Get Video Size from registry
			try
			{
				object oRwK = RegistryUtility.GetRegistry(RegistryHive.CurrentUser, Configs.RegVideoPath, Configs.RegVideoResolutionWidthKey, RegistryView.Registry64, out isX86);
				if (oRwK != null)
				{
					videoWidth = (int)oRwK;
				}
			}
			catch
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideo width from registry fail : can't get value from registry", "");
			}

			try
			{
				object oRhK = RegistryUtility.GetRegistry(RegistryHive.CurrentUser, Configs.RegVideoPath, Configs.RegVideoResolutionHeightKey, RegistryView.Registry64, out isX86);
				if (oRhK != null)
				{
					videoHeight = (int)oRhK;
				}
			}
			catch
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideo height from registry fail : can't get value from registry", "");
			}
		}

		/// <summary>
		/// Gets the video bite rate.
		/// </summary>
		/// <param name="videoBiteRate">The video bite rate.</param>
		public static void GetVideoBiteRate(ref uint videoBiteRate)
		{
			bool isX86 = true;
			object videoBiteRateValue = null;

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Registry request LowBiteRate", "");

			videoBiteRateValue = RegistryUtility.GetRegistry(RegistryHive.CurrentUser, Configs.RegVideoPath, Configs.RegVideoLowBiterate, RegistryView.Registry64, out isX86);

			if (videoBiteRateValue != null)
			{
				videoBiteRate = Convert.ToUInt32(videoBiteRateValue);

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Registry request BiteRate: " + Convert.ToUInt32(videoBiteRateValue).ToString(), "");
			}
			else
			{
				videoBiteRate = 500000;

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Registry Query Fail, reset to default BiteRate : 500000", "");
			}
		}

		/// <summary>
		/// Gets the video protocol.
		/// </summary>
		/// <returns>
		/// Default will return VideoProtocol.TCP
		/// </returns>
		public static VideoProtocol GetVideoProtocol()
		{
			// Get video protocol from registry
			try
			{
				bool isX86;
				object oRK = RegistryUtility.GetRegistry(RegistryHive.CurrentUser, Configs.RegVideoPath, Configs.RegVideoProtocolKey, RegistryView.Registry64, out isX86);

				if (oRK != null)
				{
					return (VideoProtocol)oRK;
				}
			}
			catch
			{
				return VideoProtocol.TCP;
			}

			return VideoProtocol.TCP;
		}

		public static void GetVideoKeyFrameInterval(ref uint videoKeyFrameInterval)
		{
			bool isX86 = false;

			try
			{
				object videoIntervalValue = RegistryUtility.GetRegistry(RegistryHive.CurrentUser, Configs.RegVideoPath, Configs.RegVideoIntervalKey, RegistryView.Registry64, out isX86);
				if (videoIntervalValue != null)
				{
					videoKeyFrameInterval = Convert.ToUInt32(videoIntervalValue);

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Registry request Frame Interval: " + Convert.ToUInt32(videoIntervalValue).ToString(), "");
				}
				else
				{
					videoKeyFrameInterval = 15;

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Registry Query Fail, reset to default Frame Interval : 4", "");
				}
			}
			catch
			{
				videoKeyFrameInterval = 15;
			}
		}
	}
}