﻿using AirFusion.Common;
using AirFusion.Windows;
using System;
using System.Threading;

namespace AirFusion.ConnectionManager
{
	public class FlowControlManager
	{
		#region Handle Input control

		// For controller and container
		private object _inputControllerLockObj = new object();

		private VNCController _inputController = null;

		public VNCController InputController
		{
			get
			{
				lock (_inputControllerLockObj)
				{
					return _inputController;
				}
			}

			set
			{
				lock (_inputControllerLockObj)
				{
					_inputController = value;
				}
			}
		}

		#endregion Handle Input control

		/// <summary>
		/// The _pInstanceLock lock for check FlowControlManager only one
		/// </summary>
		private static object _pInstanceLock = new object();

		/// <summary>
		/// The _pConnectionLock lock for check Connection only one
		/// </summary>
		private static object _pConnectionLock = new object();

		/// <summary>
		/// The point of flow control manager
		/// </summary>
		private static FlowControlManager _pFlowControlManager = null;

		/// <summary>
		/// The flow connect device information
		/// </summary>
		private AppDeviceInfo FlowConnectDeviceInfo;

		/// <summary>
		/// The flag is use for check has any device is connecting
		/// </summary>
		private bool IsStartNewConnection = false;

		private bool IsScreencasting = false;
		private bool IsDisconnect = false;

		public ScreenWindow _currentScreenWindow = null;
		public int m_iVideoWidth, m_iVideoHeight;
		public uint m_BiteRate;
		public uint m_iKeyFrameInterval;

		public event System.EventHandler<AppEventArgs> AppEvent;

		public Version RemoteDeviceapplicationVersion = null;
		public Version RemoteDeviceminimumTargetVersion = null;

		#region public method

		/// <summary>
		/// For avoid CheckRemoteDeviceStatusManager double create.
		/// </summary>
		public static FlowControlManager Instance
		{
			get
			{
				if (_pFlowControlManager == null)
				{
					lock (_pInstanceLock)
					{
						if (_pFlowControlManager == null) _pFlowControlManager = new FlowControlManager();
					}
				}
				return _pFlowControlManager;
			}
		}

		/// <summary>
		/// Starts the new connection.
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		public void StartNewConnection(AppDeviceInfo deviceInfo, ScreenWindow screenWindow)
		{
			if (screenWindow == null)
				return;

			//For VQ design
			if (this.FlowConnectDeviceInfo == null || !this.FlowConnectDeviceInfo.Device.InstanceID.Equals(deviceInfo.Device.InstanceID))
			{
				this.m_iVideoHeight = 0;
				this.m_iVideoWidth = 0;
				screenWindow.RemoveVQOptionToVQList();
			}

			//For change resoultion use, if the user want to change resoultion,
			//the screen height and width value is not zero.
			if (this.m_iVideoHeight != 0 || this.m_iVideoWidth != 0)
			{
				Thread.Sleep(400);
			}

			this._currentScreenWindow = screenWindow;

			lock (_pConnectionLock)
			{
				LogtoFile.TrackEnter();
				if (IsStartNewConnection == false)
				{
					this.IsStartNewConnection = true;
					this.IsScreencasting = true;
					this.IsDisconnect = false;

					//Send Event to screenwindow info step progress
					this.SendProgressStatusEvent(InitializingProgressbarArguments.MaxValue, 4);
					this.SendProgressStatusEvent(InitializingProgressbarArguments.Value, 1);

					this.FlowConnectDeviceInfo = deviceInfo;
					CheckRemoteDeviceStatusManager.Instance.Connect(deviceInfo, this);
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Already run a new connection, please try again: ", "");
				}
				LogtoFile.TrackLeave();
			}
		}

		/// <summary>
		/// The method will provided disconnect mehtod call by another code.
		/// </summary>
		public void NewConnectionDisconnect()
		{
			LogtoFile.TrackEnter();

			this.IsDisconnect = true;

			if (this.IsScreencasting)
			{
				this.ScreencastManagerDisconnected();
			}
			LogtoFile.TrackLeave();
		}

		public bool IsAcceptNewConnection()
		{
			return (!this.IsStartNewConnection);
		}

		#endregion public method

		#region private method

		/// <summary>
		/// Prevents a default instance of the <see cref="FlowControlManager"/> class from being created.
		/// </summary>
		/// When FlowControl Initial, it will add CheckRemoteDeviceStatusManager, CheckRemoteControlStatusManager
		/// and ScreencastManager event in callback function
		private FlowControlManager()
		{
			CheckRemoteDeviceStatusManager.Instance.ConnectionFlowEvent += this.ConnectStatusEventCallback;
			CheckRemoteDeviceStatusManager.Instance.ConnectStepEvent += this.ConnectStepEventCallback;

			CheckRemoteControlStatusManager.Instance.ConnectionFlowEvent += this.ConnectStatusEventCallback;
			CheckRemoteControlStatusManager.Instance.ConnectStepEvent += this.ConnectStepEventCallback;

			ScreencastManager.Instance.ConnectionFlowEvent += this.ConnectStatusEventCallback;
			ScreencastManager.Instance.ConnectStepEvent += this.ConnectStepEventCallback;

			//this.AppEvent += this._currentScreenWindow.AppEventCallback;

			this.m_iVideoHeight = 0;
			this.m_iVideoWidth = 0;
			this.m_BiteRate = 0;
			this.m_iKeyFrameInterval = 0;
		}

		private void ResetNewConnectionFlag()
		{
			this.IsStartNewConnection = false;
		}

		private void NewConnectionFinished()
		{
			LogtoFile.TrackEnter();
			this.ResetNewConnectionFlag();

			Configs.ApplicationReference.StartTimerToCheckReceiveData();
			LogtoFile.TrackLeave();
		}

		#region Handle Disconnect flow

		/// <summary>
		/// Handle All connection fail event
		/// </summary>
		/// <param name="connectionType">Type of the connection.</param>
		private void NewConnectionFailManager(ConnectionEventType connectionType)
		{
			if (connectionType == ConnectionEventType.CheckRemoteDeviceStatusSuccess ||
				connectionType == ConnectionEventType.CheckRemoteControlStatusSuccess ||
				connectionType == ConnectionEventType.ScreencastControlSuccess)
			{
				return;
			}
			else
			{
				switch (connectionType)
				{
					case ConnectionEventType.CheckRemoteDeviceStatusSuccess:
					case ConnectionEventType.CheckRemoteControlStatusSuccess:
					case ConnectionEventType.ScreencastControlSuccess:
						break;

					case ConnectionEventType.ScreencastControlFail:
						{
							this.CheckRemoteControlManagerDisconnected();
						}
						break;

					case ConnectionEventType.CheckRemoteControlStatusFail:
						{
							this.CheckRemoteDeviceStatusManagerDisconnected();
						}
						break;

					case ConnectionEventType.CheckRemoteDeviceStatusFail:
					case ConnectionEventType.None:
						{
							this.NewConnectionDisconnected();
						}
						break;

					default:
						break;
				}

				//Because new connection has been cancel, so we need to reset new connection flag
				this.ResetNewConnectionFlag();
			}
		}

		/// <summary>
		/// When Screencast Manager has been disconnected,
		/// need to call previous step (CheckRemoteControlStatusManager) to disconnected
		/// When Previous step has been disconnected, need to callback to Fail control Manager to handle it
		/// </summary>
		private void ScreencastManagerDisconnected()
		{
			LogtoFile.TrackEnter();
			CheckRemoteControlStatusManager.Instance.Disconnect(this.FlowConnectDeviceInfo, false);
			this.NewConnectionFailManager(ConnectionEventType.CheckRemoteControlStatusFail);
			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// When Control Status Manager has been disconnected,
		/// need to call previous step (CheckRemoteDeviceStatusManager) to disconnected
		/// When Previous step has been disconnected, need to callback to Fail control Manager to handle it
		/// </summary>
		private void CheckRemoteControlManagerDisconnected()
		{
			LogtoFile.TrackEnter();
			CheckRemoteDeviceStatusManager.Instance.Disconnect(FlowConnectDeviceInfo, false);
			this.NewConnectionFailManager(ConnectionEventType.CheckRemoteDeviceStatusFail);
			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// When Device Status Manager has been disconnected, need to callback to Fail control Manager to handle it
		/// </summary>
		private void CheckRemoteDeviceStatusManagerDisconnected()
		{
			LogtoFile.TrackEnter();
			this.NewConnectionFailManager(ConnectionEventType.None);
			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// News the connection has been disconnected.
		/// </summary>
		private void NewConnectionDisconnected()
		{
			LogtoFile.TrackEnter();

			this.ResetNewConnectionFlag();

			this.IsScreencasting = false;
			this.IsDisconnect = false;

			this.SendProgressStatusEvent(InitializingProgressbarArguments.Value, 0);

			LogtoFile.TrackLeave();
		}

		#endregion Handle Disconnect flow

		#endregion private method

		#region Event Handle

		private void CheckRemoteDeviceStatusManagerErrorEventHandle(ConnectDeviceStatusType Errorstatus)
		{
			LogtoFile.TrackEnter();

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckConnectionStatusType: " + Errorstatus.GetStringValue(), "");
			try
			{
				switch (Errorstatus)
				{
					case ConnectDeviceStatusType.WakeupFail:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeWakeUpFail));
						}
						break;

					case ConnectDeviceStatusType.ConfirmIPFail:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.DandelionReconfirmIPFail));
						}
						break;

					case ConnectDeviceStatusType.DandelionFail:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.DandelionGetIPFail));
						}
						break;

					case ConnectDeviceStatusType.VersionNotMatch:
						{
							if (RemoteDeviceapplicationVersion != null && RemoteDeviceminimumTargetVersion != null)
							{
								Version applicationVersion = new Version(Configs.ApplicationReference.ApplicationVersion);

								Version minimumTargetVersion = new Version(Configs.ApplicationReference.MinimumTargetVersion);

								Configs.ApplicationReference.ShowUpdateMessage(applicationVersion, minimumTargetVersion, this.RemoteDeviceapplicationVersion, this.RemoteDeviceminimumTargetVersion);

								//Need to Update Android Deivice
								//if (minimumTargetVersion > RemoteDeviceminimumTargetVersion)
								//{
								//	string strAndroidUpdate = "Please update your AcerEXTEND on your mobile";
								//	string strAndroidUpdateDetail = "GoToUpdate";
								//	CustomMessageBox androidUpdateMessageBox = new CustomMessageBox(new AirFusion.Windows.CustomMessageBox.CustomMessageBoxStructInfo(
								//													strAndroidUpdate, CustomMessageBox.CustomImageIcon.Update, strAndroidUpdateDetail));
								//	androidUpdateMessageBox.ShowDialog();

								//	if (androidUpdateMessageBox.MessageBoxResult == System.Windows.MessageBoxResult.OK)
								//	{
 
								//	}
								//}
								//else if (applicationVersion < RemoteDeviceminimumTargetVersion) // Need to Update Windows Version
								//{
								//	string strAndroidUpdate = "Please update your AcerEXTEND on your PC";
								//	string strAndroidUpdateDetail = "GoToUpdate";
								//	CustomMessageBox WindowsUpdateMessageBox = new CustomMessageBox(new AirFusion.Windows.CustomMessageBox.CustomMessageBoxStructInfo(
								//													strAndroidUpdate, CustomMessageBox.CustomImageIcon.Update, strAndroidUpdateDetail));
								//	WindowsUpdateMessageBox.ShowDialog();

								//	if (WindowsUpdateMessageBox.MessageBoxResult == System.Windows.MessageBoxResult.OK)
								//	{
 
								//	}
								//}
							}
						}
						break;

					case ConnectDeviceStatusType.CheckDeviceStatusUnPair:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.PairAgain, this.FlowConnectDeviceInfo.Device));
						}
						break;

					case ConnectDeviceStatusType.CheckDeviceStatusTimeout:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeGetDeviceStatusTimeout));
						}
						break;

					case ConnectDeviceStatusType.CheckDeviceStatusFail:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeGetDeviceStatusFail));
						}
						break;

					case ConnectDeviceStatusType.GetDeviceInfoTimeout:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeGetDeviceInformationTimeout));
						}
						break;

					case ConnectDeviceStatusType.GetDeviceInfoFail:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeGetDeviceInformationFail));
						}
						break;

					default:
						{ }
						break;
				}
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		private void CheckRemoteControlManagerErrorEventHandle(ConnectControlStatusType Errorstatus)
		{
			LogtoFile.TrackEnter();

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceControl Type: " + Errorstatus.GetStringValue(), "");
			try
			{
				switch (Errorstatus)
				{
					case ConnectControlStatusType.InitialInputControlFail:
					case ConnectControlStatusType.InputControlConnectFail:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeConnectToVNCFail));
						}
						break;

					case ConnectControlStatusType.InputDaemonNotRunning:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InputDaemonNotRunning));
						}
						break;

					case ConnectControlStatusType.CheckInputDaemonTimeout:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeInputControlTimeout));
						}
						break;

					case ConnectControlStatusType.CheckInputDaemonWebException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeInputControlWebServiceException));
						}
						break;

					case ConnectControlStatusType.CheckInputDaemonException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeInputControlWebServiceException));
						}
						break;

					case ConnectControlStatusType.ScreencastDaemonNotRunning:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeScreenCastDaemonFail));
						}
						break;

					case ConnectControlStatusType.CheckScreencastDaemonTimeout:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeScreencastDaemonTimeout));
						}
						break;

					case ConnectControlStatusType.CheckScreencastDaemonWebException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeScreencastDaemonWebServiceException));
						}
						break;

					case ConnectControlStatusType.CheckScreencastDaemonException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeScreencastDaemonWebServiceException));
						}
						break;					
					case ConnectControlStatusType.ADBRemoteDeviceNotFound:
					case ConnectControlStatusType.ADBPullInputDaemonFail:
					case ConnectControlStatusType.ADBPushInputDaemonFail:
					case ConnectControlStatusType.ADBShellSetPrivilegeInputDaemonFail:
					case ConnectControlStatusType.LaunchInputDaemonByADBException:
					case ConnectControlStatusType.ADBPullScreencastDaemonFail:
					case ConnectControlStatusType.ADBPushScreencastDaemonFail:
					case ConnectControlStatusType.ADBShellSetPrivilegeScreencastDaemonFail:
					case ConnectControlStatusType.LaunchScreencastDaemonByADBException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.NonAcerFlowFailStatus, Errorstatus));
						}
						break;

					default:
						{ }
						break;
				}
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		private void ScreencastManagerErrorEventHandle(ScreencastStatusType Errorstatus)
		{
			LogtoFile.TrackEnter();

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "ScreencastManagerErrorEventHandle Type: " + Errorstatus.GetStringValue(), "");
			try
			{
				switch (Errorstatus)
				{
					case ScreencastStatusType.ScreencastControlTimeout:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.PlayScreencastTimeout));
						}
						break;

					case ScreencastStatusType.ScreencastControlWebServiceException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.PlayScreencastError));
						}
						break;

					case ScreencastStatusType.ScreencastControlException:
						{
							this.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.PlayScreencastError));
						}
						break;

					default:
						{ }
						break;
				}
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		private async void ConnectStatusEventCallback(object sender, ConnectEventArgs e)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get ConnectionEventType: " + e.EventType.GetStringValue(), "");

			switch (e.EventType)
			{
				#region Handle CheckRemoteDeviceStauts Manager Event

				//When check remote device status success, need to do next step "CheckRemoteControlStatus" is OK
				case ConnectionEventType.CheckRemoteDeviceStatusSuccess:
					{
						if (!this.IsDisconnect)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "ConnectionEventType.CheckRemoteDeviceStatusSuccess", "");

							//Send Event to screenwindow info step progress
							this.SendProgressStatusEvent();

							CheckRemoteControlStatusManager.Instance.Connect(this.FlowConnectDeviceInfo, this);
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceStatusSuccess but connect has been disconnect, should not send event to next filter", "");
						}
					}
					break;

				//Handle Check Remote Device Status Fail case
				case ConnectionEventType.CheckRemoteDeviceStatusFail:
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "ConnectionEventType.CheckRemoteDeviceStatusFail", "");

						if (null != e.Data && e.Data is ConnectDeviceStatusType)
						{
							ConnectDeviceStatusType Errorstatus = (ConnectDeviceStatusType)e.Data;

							CheckRemoteDeviceStatusManagerErrorEventHandle(Errorstatus);

							this.CheckRemoteDeviceStatusManagerDisconnected();
						}
						else
						{
							if (null == e.Data)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceStatusFail but CheckConnectionStatusType is null", "");
							}
							else
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceStatusFail but CheckConnectionStatusType is format error", "");
							}
						}
					}
					break;

				#endregion Handle CheckRemoteDeviceStauts Manager Event

				#region Handel CheckRemoteControlStatus Manager Event

				//When check remote device status success, need to do next step "Screencast connect" is OK
				case ConnectionEventType.CheckRemoteControlStatusSuccess:
					{
						if (!this.IsDisconnect)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "ConnectionEventType.CheckRemoteControlStatusSuccess", "");

							//Send Event to screenwindow info step progress
							this.SendProgressStatusEvent();

							ScreencastManager.Instance.Connect(this.FlowConnectDeviceInfo, this);
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteControlStatusSuccess but connect has been disconnect, should not send event to next filter", "");
						}
					}
					break;

				case ConnectionEventType.CheckRemoteControlStatusFail:
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "ConnectionEventType.CheckRemoteControlStatusFail", "");

						if (null != e.Data && e.Data is ConnectControlStatusType)
						{
							ConnectControlStatusType Errorstatus = (ConnectControlStatusType)e.Data;

							CheckRemoteControlManagerErrorEventHandle(Errorstatus);

							this.CheckRemoteControlManagerDisconnected();
						}
						else
						{
							if (null == e.Data)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteControlStatusFail but ConnectControlStatusType is null", "");
							}
							else
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteControlStatusFail but ConnectControlStatusType is format error", "");
							}
						}
					}
					break;

				#endregion Handel CheckRemoteControlStatus Manager Event

				#region Handle Screencast Manager Event

				case ConnectionEventType.ScreencastControlSuccess:
					{
						if (!this.IsDisconnect)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "ConnectionEventType.ScreencastControlSuccess", "");

							//Send Event to screenwindow info step progress
							this.SendProgressStatusEvent();

							this.NewConnectionFinished();
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "ScreencastControlSuccess but connect has been disconnect, should not send event to next filter", "");
						}
					}
					break;

				case ConnectionEventType.ScreencastControlFail:
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "ConnectionEventType.ScreencastControlFail", "");

						if (null != e.Data && e.Data is ScreencastStatusType)
						{
							ScreencastStatusType Errorstatus = (ScreencastStatusType)e.Data;

							ScreencastManagerErrorEventHandle(Errorstatus);

							this.ScreencastManagerDisconnected();
						}
						else
						{
							if (null == e.Data)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "ScreencastControlFail but ScreencastStatusType is null", "");
							}
							else
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "ScreencastControlFail but ScreencastStatusType is format error", "");
							}
						}
					}
					break;

				#endregion Handle Screencast Manager Event

				default:
					break;
			}
		}

		private async void ConnectStepEventCallback(object sender, AppEventArgs e)
		{
			this.SendAppEventAsync(e);
		}

		private void SendProgressStatusEvent()
		{
			this.SendAppEventAsync(new AppEventArgs(AppEventType.NextProgress));
		}

		private void SendProgressStatusEvent(InitializingProgressbarArguments key, int value)
		{
			this.SendAppEventAsync(new AppEventArgs(AppEventType.NextProgress, key, value));
		}

		public void SendAppEventAsync(AppEventArgs args)
		{
			try
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Ready to send AppEvent: " + args.EventType, "");

				if (null != this.AppEvent)
				{
					var receivers = this.AppEvent.GetInvocationList();
					foreach (EventHandler<AppEventArgs> receiver in receivers)
					{
						receiver.BeginInvoke(this, args, null, null);
						//receiver.Invoke(this, args);
					}
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this.AppEvent is null", "");
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		#endregion Event Handle
	}
}
