﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Interop;

namespace AirFusion
{
	/// <summary>
	/// Interaction logic for TrayIconWindow.xaml
	/// </summary>
	public partial class TrayIconWindow : Window
	{
		#region Members

		private HwndSource _source;

		#endregion Members

		#region Events

		public event EventHandler<WindowEventArgs> WindowEvent;

		#endregion Events

		#region Properties

		public ICommand DoubleClickTrayCommand { get; private set; }

		#endregion Properties

		#region Constructor & Dispose

		public TrayIconWindow()
		{
			this.DoubleClickTrayCommand = new DelegateCommand(this.TrayDoubleClicked, null);
			InitializeComponent();
			InitializeUIComponent();

			this.DataContext = this;
		}

		#endregion Constructor & Dispose

		#region Window-related Event Handlers

		private void Window_Loaded(object sender, RoutedEventArgs e)
		{
			_source = HwndSource.FromHwnd(new WindowInteropHelper(this).Handle);
			_source.AddHook(new HwndSourceHook(WndProc));
		}

		private void Window_StateChanged(object sender, EventArgs e)
		{
			if (this.WindowState != WindowState.Minimized)
			{
				this.WindowState = WindowState.Minimized;
			}
		}

		private void Window_Closed(object sender, EventArgs e)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start", "");

			try
			{
				if (null != this._source)
				{
					this._source.RemoveHook(this.WndProc);
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "End", "");
		}

		#endregion Window-related Event Handlers

		#region Private Methods

		private void InitializeUIComponent()
		{
			this.Dispatcher.Invoke(delegate()
			{
				try
				{
					this.MenuOpen.ItemsSource = null;
					this.MenuOpen.ItemsSource = Configs.ApplicationReference.PairedDevices;
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				}
			});
		}

		private void SendWindowEventAsync(object sender, WindowEventArgs eventArgs)
		{
			try
			{
				if (null != this.WindowEvent)
				{
					// To send event by using different thread, not UI thread
					//this.WindowEvent.BeginInvoke(this, eventArgs, null, null);
					this.WindowEvent.Invoke(this, eventArgs);
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		private void TrayDoubleClicked(object parameter)
		{
			try
			{
				Configs.ApplicationReference.TrayDoubleClicked();
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		#endregion Private Methods

		#region Callbacks

		private IntPtr WndProc(IntPtr hwnd, int msg, IntPtr wParam, IntPtr lParam, ref bool handled)
		{
			System.Windows.Forms.Message m = System.Windows.Forms.Message.Create(hwnd, msg, wParam, lParam);

			if (m.Msg == Win32Api.WM_ShowUp)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get Win32Api.WM_ShowUp from seconed instance, to send event", "");
				this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.ExistedLaunch));
			}
			else if (m.Msg == Win32Api.WM_BADNETWORK)
			{
				// If the original network status is good, to send "BadNetwork" event once receiving message from source filter
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get Win32Api.WM_BADNETWORK from source filter, to send event", "");
				Configs.ApplicationReference.IsNetworkGood = false;
			}
			else if (m.Msg == Win32Api.WM_GOODNETWORK)
			{
				// If the original network status is bad, to send "GoodNetwork" event once receiving message from source filter
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get Win32Api.WM_GOODNETWORK from source filter, to send event", "");
				Configs.ApplicationReference.IsNetworkGood = true;
			}
			else if (m.Msg == Win32Api.WM_ForceExit)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get Win32Api.WM_ForceExit, to send event", "");
				this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.Exit));
			}
			else if (m.Msg == Win32Api.WM_CHECKRECEIVEDATA)
			{
				// Set the signal of starting timer
				Configs.ApplicationReference.SetReceiveDataEvent();
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get Win32Api.WM_CHECKRECEIVEDATA, to send event", "");
				this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.CheckReceiveData));
			}
			else if (m.Msg == Win32Api.WM_TCPDisconnected)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get Win32Api.WM_TCPDisconnected, to send event", "");
				this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.TCPDisconnected));
			}

			return IntPtr.Zero;
		}

		#endregion Callbacks

		#region UI Event Callbacks

		private void MenuOpen_Click(object sender, RoutedEventArgs e)
		{
			try
			{
				if (null != sender)
				{
					AppDeviceInfo selectedItem = (sender as MenuItem).DataContext as AppDeviceInfo;
					this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.RestartScreenCasting, selectedItem, true));
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}

			return;
		}

		private void MenuAddDevice_Click(object sender, RoutedEventArgs e)
		{
			this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.ShowDeviceManager, null, DeviceManagerTab.DeviceManager_Pairing));
		}

		private void MenuShowMyDevice_Click(object sender, RoutedEventArgs e)
		{
			this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.ShowDeviceManager, null, DeviceManagerTab.DeviceManager_Paired));
		}

		private void MenuDownloads_Click(object sender, RoutedEventArgs e)
		{
			try
			{
				if (!System.IO.Directory.Exists(Configs.DownloadFileFolderPath))
				{
					System.IO.Directory.CreateDirectory(Configs.DownloadFileFolderPath);
				}

				Process.Start(Configs.DownloadFileFolderPath);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		private void MenuAbout_Click(object sender, RoutedEventArgs e)
		{
		}

		private void MenuCheckForUpdates_Click(object sender, RoutedEventArgs e)
		{
			Configs.ApplicationReference.UserCheckforUpdatesAsync();
			//UpgradeTool.ExitValue CheckUpdate = UpgradeTool.SelfCheckUpdte(UpgradeTool.LaunchType.UserCheckUpdate);
			////Need to modify, wait for UI design general message dialog
			//if (CheckUpdate == UpgradeTool.ExitValue.NONetWork)
			//	MessageBox.Show(Application.Current.TryFindResource("txtUpgradeToolNONetWork") as String);
			//else if(CheckUpdate == UpgradeTool.ExitValue.NONewVersion)
			//	MessageBox.Show(Application.Current.TryFindResource("txtUpgradeToolNONewVersion") as String);
		}

		private void MenuExit_Click(object sender, RoutedEventArgs e)
		{
			//Johnson 0416 Add for debug Close window deadlock issue
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Tray Icon MenuExit Click", "");

			this.SendWindowEventAsync(this, new WindowEventArgs(WindowEventType.Exit));
		}

		#endregion UI Event Callbacks
	}
}