﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Management;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Common.Wmi
{
	/// <summary>
	/// Represents a template of the WMI reader of a specific WMI class.
	/// </summary>
	/// <typeparam name="TClass">The type of the WMI class.</typeparam>
	/// <example>
	/// <![CDATA[
	/// foreach (WmiComputerSystem computerSystem in WmiReader<WmiComputerSystem>.Get())
	/// {
	///    ...
	/// }
	/// ]]>
	/// </example>
	public class WmiReader<TClass>
		where TClass : class, new()
	{
		/// <summary>
		/// Returns an enumerator that iterates through the instances.
		/// </summary>
		/// <returns></returns>
		/// <exception cref="System.Exception" />
		/// <exception cref="System.ArgumentException" />
		public static IEnumerable<TClass> Get()
		{
			try
			{
				StaticLogger.TrackEnter();

				string className = null;

				// find first ParameterValue value with string type for the class name
				foreach (ParameterValue parameterValue in typeof(TClass).GetCustomAttributes(typeof(ParameterValue), true).OfType<ParameterValue>())
				{
					if (parameterValue.Value is string)
					{
						className = (parameterValue.Value as string).Trim();
						break;
					}
				}

				if (className == null)
				{
					throw new Exception("The class '" + typeof(TClass).Name + "' doesn't carry a ParameterValue attribute that contains the class name.");
				}

				// check parameters
				if (string.IsNullOrWhiteSpace(className))
				{
					throw new ArgumentException("Please specify the class name.");
				}
				if (className.Any(x => char.IsWhiteSpace(x)))
				{
					throw new ArgumentException("The class name should not contain any space.");
				}

				string queryString = "SELECT * FROM " + className;
				StaticLogger.LogInfo("WMI query string: " + queryString);
				using (ManagementObjectSearcher searcher = new ManagementObjectSearcher(new ManagementScope("root\\CIMV2"), new ObjectQuery(queryString)))
				{
					ManagementObjectCollection collection = null;
					try
					{
						collection = searcher.Get();
					}
					catch (ManagementException ex)
					{
						StaticLogger.LogError("Failed getting the collection: " + ex.Message + " (ErrorCode=" + ex.ErrorCode + ")");
						throw;
					}
					if (collection == null)
					{
						yield break;
					}
					else
					{
						foreach (ManagementObject instance in collection)
						{
							TClass obj = new TClass();
							Fill(instance, ref obj);
							yield return obj;
						}
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Fills the values into TClass object from an ManagementObject instance.
		/// </summary>
		/// <param name="instance"></param>
		/// <param name="obj"></param>
		protected static void Fill(ManagementObject instance, ref TClass obj)
		{
			try
			{
				StaticLogger.TrackEnter();

				foreach (var property in obj.GetType().GetProperties())
				{
					foreach (ParameterValue parameterValue in property.GetCustomAttributes(typeof(ParameterValue), true).OfType<ParameterValue>())
					{
						if (parameterValue != null && parameterValue.Value is string)
						{
							string propertyName = parameterValue.Value as string;
							Type propertyType = property.PropertyType;
							if (propertyType.IsGenericType && propertyType.GetGenericTypeDefinition() == typeof(Nullable<>))
							{
								propertyType = Nullable.GetUnderlyingType(propertyType);
							}

							try
							{
								if (instance.GetPropertyValue(propertyName) != null)
								{
									Type objectType = instance[propertyName].GetType();
									TypeConverterTypeAttribute typeConverterTypeAttribute = property.GetCustomAttributes(typeof(TypeConverterTypeAttribute), false).OfType<TypeConverterTypeAttribute>().FirstOrDefault();

									// if there's a converter
									if (typeConverterTypeAttribute != null && typeConverterTypeAttribute.ConverterType != null)
									{
										try
										{
											TypeConverter converter = Activator.CreateInstance(typeConverterTypeAttribute.ConverterType) as TypeConverter;
											if (!converter.CanConvertFrom(objectType))
											{
												throw new Exception("The converter is not able to converter from '" + objectType.Name + "' type.");
											}
											property.SetValue(obj, converter.ConvertFrom(instance[propertyName]));
										}
										catch (Exception ex)
										{
											throw new Exception("Failed converting on property '" + propertyName + "'.", ex);
										}
									}
									// compare type and set the value if type matches
									else if (propertyType.Equals(objectType))
									{
										property.SetValue(obj, instance[propertyName]);
									}
									else
									{
										throw new Exception("The type of '" + propertyName + "' is inconsistent (" + propertyType.Name + ", " + objectType.Name + ").");
									}
									continue;
								}
							}
							catch (Exception ex)
							{
								StaticLogger.LogWarn("Failed fetching property '" + propertyName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						}
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
	}
}
