﻿using AirFusion.Manager.Notification;
using AirFusion.Windows;
using System;
using System.Collections.Concurrent;
using System.Diagnostics;
using System.Drawing;
using System.Threading;
using System.Windows;

namespace AirFusionCustomToast
{
	public class GeneralNotificationManager
	{
		#region Members

		private Application _app = null;
		private bool _isMuteAllNotification = false;
		private SynchronizationContext _sc = null;

		#endregion Members

		#region Properties

		public NotificationManager NotificationManager { get; internal set; }

		#endregion Properties

		#region Events

		public event EventHandler<ToastEventArgs> NotificationEvent;

		#endregion Events

		#region Constructors & Dispose

		public GeneralNotificationManager(NotificationManager notificationManager)
		{
			if (notificationManager != null)
			{
				this.NotificationManager = notificationManager;
				this.NotificationManager.NotificationRemoved += OnNotificationRemoved;
				this.NotificationManager.NotificationReceived += OnNotificationReceived;
				this.NotificationManager.NotificationIgnoreTimeOuted += NotificationIgnoreTimeOutedHandler;
			}

			if (SynchronizationContext.Current == null)
			{
				throw new InvalidOperationException("Please initialize this instance by main thread.");
			}

			this._sc = SynchronizationContext.Current;
		}

		private void NotificationIgnoreTimeOutedHandler(object sender, AirFusion.Manager.Notification.NotificationManager.AFNotificationIgnoreTimeOutedEventArgs args)
		{
			if (args.RetryTime < 2)
				this.NotificationManager.NotificationIgnore(args.PackageName, args.DeviceInstanceID, args.NotificationKey, args.RetryTime + 1);
			else
			{
				this._sc.Post(delegate(object param)
				{
					string GuestAccountContent = (string)System.Windows.Application.Current.TryFindResource("txtGeneralAlertDetail"); //"Failed operation";
					string ButtonContent = (string)System.Windows.Application.Current.TryFindResource("txtOK");
					CustomMessageBox guestMessageBox = new CustomMessageBox(new AirFusion.Windows.CustomMessageBox.CustomMessageBoxStructInfo(
																	GuestAccountContent, CustomMessageBox.CustomImageIcon.Warring, ButtonContent));

					guestMessageBox.ShowDialog();
				}, null);
			}
		}

		private void OnNotificationRemoved(object sender, NotificationManager.AFNotificationItemRemoveEventArgs args)
		{
			// do nothing now
		}

		private void OnNotificationReceived(object sender, NotificationManager.AFNotificationItemUpdateEventArgs args)
		{
			if (!this._isMuteAllNotification)
			{
				string notificationKey = args.NotificationKey;
				string deviceName = args.DeviceName;
				ToastItem item = args.Toast;

				this.CreateAndShowToast(notificationKey, deviceName, item);
			}
		}

		/// <summary>
		/// Mutes all notifications.
		/// </summary>
		/// <param name="isMuteAllNotification">mutes all notifications if true</param>
		public void MuteAllNotification(bool isMuteAllNotification)
		{
			this._isMuteAllNotification = isMuteAllNotification;
		}

		#endregion Constructors & Dispose

		#region Private Methods

		private void SendEventAsync(object sender, ToastEventArgs eventArgs)
		{
			try
			{
				if (null != this.NotificationEvent)
				{
					this.NotificationEvent(sender, eventArgs);
				}
			}
			catch { }
		}

		#endregion Private Methods

		#region Public Methods

		/// <summary>
		/// Creates and show the specified toast.
		/// </summary>
		/// <param name="key"></param>
		/// <param name="deviceName"></param>
		/// <param name="item"></param>
		public void CreateAndShowToast(string key, string deviceName, AirFusion.Manager.Notification.ToastItem item)
		{
			Debug.Assert(this._sc != null);

			if (string.IsNullOrEmpty(key) || string.IsNullOrEmpty(deviceName) || item == null)
				return;

			this._sc.Post(delegate(object param)
			{
				GeneralNotificationUIElement notificationItem = new GeneralNotificationUIElement();

				// Basic information
				{
					//inNofication.Launchable = item.ActivityPopup.HasValue ? item.ActivityPopup.Value : false;
					notificationItem.Removable = item.Removable;
					notificationItem.Title = item.ToastTitle;
					notificationItem.AppName = item.AppName;
					notificationItem.NotificationKey = key;
					notificationItem.DeviceInstanceID = item.DeviceInstanceId.Value;
					notificationItem.Type = item.Type;
					notificationItem.isMediaStyle = item.isMediaStyle;
					notificationItem.IsAllowMute = item.isAllowMute;
				}

				// Content
				{
					if (item.ToastContent != null)
					{
						foreach (string content in item.ToastContent)
						{
							notificationItem.Content += content + "\r\n";
						}
					}
				}

				// Info
				{
					notificationItem.Info = deviceName + " at ";
					notificationItem.Info += DateTime.Now.ToString("HH:mm");
				}

				// Quick Actions
				{
					if (item.QAItems != null && item.QAItems.Count > 0)
					{
						foreach (AirFusion.Manager.Notification.QuickActionItem qa in item.QAItems)
						{
							notificationItem.QAItemList.Add(qa);
						}
					}
				}

				// Icon
				{
					notificationItem.BitmapIcon = null;

					if (!string.IsNullOrEmpty(item.IconPath) && System.IO.File.Exists(item.IconPath))
					{
						try
						{
							Bitmap bitmapIcon = new Bitmap(System.Drawing.Image.FromFile(item.IconPath));
							System.Windows.Media.Imaging.BitmapSource bitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
										bitmapIcon.GetHbitmap(),
										IntPtr.Zero,
										System.Windows.Int32Rect.Empty,
										System.Windows.Media.Imaging.BitmapSizeOptions.FromWidthAndHeight(bitmapIcon.Width, bitmapIcon.Height));

							bitmapSource.Freeze();
							notificationItem.BitmapIcon = bitmapSource;
						}
						catch (Exception ex)
						{ }
					}
				}

				// Local Launch Handler
				{
					notificationItem.LocalLaunchHandler = item.LocalLaunchItem;
				}

				GeneralToastPopUp toast = new GeneralToastPopUp(notificationItem);

				#region register events

				toast.CustomToastActivated += delegate(object sender, GeneralToastPopUp.CustomToastEventRawArgs args)
				{
					this.SendEventAsync(this, new ToastEventArgs(ToastEventType.ToastActivated, ToastIntent.ShowScreenCasting, args.NotificationKey, args.DeviceId));
				};

				toast.CustomToastLaunched += delegate(object sender, GeneralToastPopUp.CustomToastEventRawArgs args)
				{
					this.NotificationManager.NotificationLaunch(args.NotificationKey);
					this.SendEventAsync(this, new ToastEventArgs(ToastEventType.ToastActivated, ToastIntent.ShowScreenCasting, args.NotificationKey, args.DeviceId));
				};

				toast.CustomToastDismissed += delegate(object sender, GeneralToastPopUp.CustomToastEventRawArgs args)
				{
					this.NotificationManager.NotificaitonRemove(args.NotificationKey);
				};

				toast.CustomToastWindowClosed += delegate(object sender, GeneralToastPopUp.CustomToastEventRawArgs args)
				{
					foreach (Window window in System.Windows.Application.Current.Windows)
					{
						if (window == null || !window.GetType().Equals(typeof(GeneralToastPopUp))) continue;

						if (window.Title.Equals(args.NotificationKey))
						{
							return;
						}
					}

					this.NotificationManager.RemoveItemOnLocalDatabase(args.NotificationKey);
				};

				toast.CustomToastIgnore += delegate(object sender, GeneralToastPopUp.CustomToastEventRawArgs args)
				{
					this.NotificationManager.NotificationIgnore(string.Empty, Guid.Empty, args.NotificationKey, 0);
				};

				toast.CustomToastQuickActionActivated += delegate(object sender, GeneralToastPopUp.CustomToastQuickActionEventArgs args)
				{
					this.NotificationManager.QuickActionLaunch(args.NotificationKey, args.QuickActionId);
					if (args.QuickActionActivityPopup)
					{
						this.SendEventAsync(this, new ToastEventArgs(ToastEventType.ToastActivated, ToastIntent.ShowScreenCasting, args.NotificationKey, args.DeviceId));
					}
				};

				toast.CustomToastLocalLaunch += delegate(object sender, GeneralToastPopUp.CustomToastLocalLaunchEventArgs args)
				{
					AirFusion.Manager.Notification.LocalLaunch localLaunchHandler = args.LocalLaunchHandler;
					if (localLaunchHandler != null)
					{
						localLaunchHandler.DoAction(args.LocalLaunchIndex);
					}
				};

				#endregion register events

				toast.Show();
			}, null);
		}

		/// <summary>
		/// Creates show the specified snapshot toast.
		/// </summary>
		/// <param name="title">The notification title.</param>
		/// <param name="content">The notification content.</param>
		/// <param name="icon">The icon bitmap.</param>
		/// <param name="filePath">The file path.</param>
		public void CreateAndShowToastOfSnapshot(string title, string content, System.Windows.Media.Imaging.BitmapSource icon, string filePath)
		{
			if (string.IsNullOrEmpty(title))
			{
				throw new ArgumentException("Please specify 'title'.", title);
			}

			if (icon == null)
			{
				throw new ArgumentNullException("null");
			}

			Debug.Assert(this._sc != null);
			this._sc.Post(delegate(object param)
			{
				try
				{
					GeneralNotificationUIElement toastElement = new GeneralNotificationUIElement()
					{
						Title = title,
						Content = content,
						BitmapIcon = icon,
					};

					GeneralToastPopUp toast = new GeneralToastPopUp(toastElement, filePath);
					toast.Show();
				}
				catch (Exception ex)
				{
					string str = ex.ToString();
				}
			}, null);
		}

		public void ConfigureNotificationStatus(string deviceId, DeviceIgnoreStatus status)
		{
			Debug.Assert(this.NotificationManager != null);
			this.NotificationManager.ConfigureDeviceNotificationStatus(deviceId, status);
		}

		#endregion Public Methods
	}
}