﻿using AirFusion.Common;
using AirFusion.Manager.IncomingCall;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows;
using System.Windows.Media.Imaging;

namespace AirFusionCustomToast
{
	public class PhoneCallToastManager : IDisposable
	{
		internal struct PairedDevice
		{
			internal bool isPhoneCallToastEnabled;
			internal ToastPopUp phoneCallToast;
			internal ICCallReceivedEventArgs phoneCallProperties;
		}

		private bool _isDisposed = false;
		private bool _isAnsweringPhoneCall = false;
		private Guid _currentActivatedDeviceID = Guid.Empty;
		private IncomingCallManager _incomingCallManager;
		private ConcurrentDictionary<Guid, PairedDevice> _pairedDevices;

		public PhoneCallToastManager(IncomingCallManager incomingCallManager)
		{
			this._incomingCallManager = incomingCallManager;
			this._pairedDevices = new ConcurrentDictionary<Guid, PairedDevice>();

			this._incomingCallManager.CallReceived += new ICCallReceivedEventHandler(PhoneCallReceivedHandler);
			this._incomingCallManager.StateChanged += new ICStateChangedEventHandler(PhoneCallStateChangedHandler);
			this._incomingCallManager.Enable(true);
		}

		public void EnablePhoneCallToast(Guid deviceID, bool isEnabled)
		{
			StaticLogger.LogInfo("[###] EnablePhoneCallToast =>");
			StaticLogger.LogInfo("[###] Current device: " + deviceID.ToString());
			StaticLogger.LogInfo("[###] Phone call toast enabled: " + isEnabled);

			PairedDevice pairedDevice = new PairedDevice()
			{
				isPhoneCallToastEnabled = isEnabled,
				phoneCallToast = null,
				phoneCallProperties = null,
			};

			if (false == this._pairedDevices.ContainsKey(deviceID))
			{
				// New device => add it to list
				this._pairedDevices.AddOrUpdate(deviceID, (key) => pairedDevice, (key, oldValue) => pairedDevice);
			}
			else
			{
				// The device is exist
				pairedDevice.phoneCallToast = this._pairedDevices[deviceID].phoneCallToast;
				this._pairedDevices.AddOrUpdate(deviceID, (key) => pairedDevice, (key, oldValue) => pairedDevice);
			}
		}

		public void Dispose()
		{
			this.Dispose(true);
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if (this._isDisposed) return;

			if (disposing)
			{
				// Free any other managed objects here
				this._incomingCallManager = null;
				this.ClearPairedDeviceList();
				this._pairedDevices = null;

				this._incomingCallManager.CallReceived -= new ICCallReceivedEventHandler(PhoneCallReceivedHandler);
				this._incomingCallManager.StateChanged -= new ICStateChangedEventHandler(PhoneCallStateChangedHandler);
				this._incomingCallManager.Enable(false);
			}

			this._isDisposed = true;
		}

		private void ClearPairedDeviceList()
		{
			foreach (PairedDevice pairedDevice in this._pairedDevices.Values)
				this.FullyCloseToast(pairedDevice.phoneCallToast);

			this._pairedDevices.Clear();
		}

		private void PhoneCallReceivedHandler(object sender, ICCallReceivedEventArgs e)
		{
			StaticLogger.LogInfo("[###] Receive phone call from" + e.Device.InstanceID);

			if (this._pairedDevices.ContainsKey(e.Device.InstanceID))
			{
				if (this._pairedDevices[e.Device.InstanceID].isPhoneCallToastEnabled == true)
				{
					ToastPopUp toast = toast = this.CreatePhoneCallToast(e);

					if (toast != null)
					{
						PairedDevice pairedDevice = new PairedDevice()
						{
							isPhoneCallToastEnabled = true,
							phoneCallToast = toast,
							phoneCallProperties = e,
						};

						this._pairedDevices.AddOrUpdate(e.Device.InstanceID, (key) => pairedDevice, (key, oldValue) => pairedDevice);
					}
				}
				else
				{
					StaticLogger.LogInfo("[###] Device, " + e.Device.InstanceID + ", is not enabled phone call toast.");
				}
			}
			else
			{
				StaticLogger.LogInfo("[###] _pairedDevices list doesn't contain the device => " + e.Device.InstanceID);
			}
		}

		private ToastPopUp CreatePhoneCallToast(ICCallReceivedEventArgs e)
		{
			ToastPopUp toast = null;

			try
			{
				System.Windows.Application.Current.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Normal, new Action(
					() =>
					{
						BitmapSource caller_photo = null;
						if (e.CallerPhoto == null)
						{
							string imgSource = @"pack://application:,,,/AirFusionCustomToast;component/Resources/Caller.png";
							caller_photo = new System.Windows.Media.Imaging.BitmapImage(new Uri(imgSource, UriKind.Absolute));
						}
						else
						{
							caller_photo = ConvertBitmap(e.CallerPhoto);
						}

						ToastUIElement toastUIElement = new ToastUIElement
						{
							Title = e.PhoneNumber + " " + e.CallerName,
							ContentLine1 = "Incoming call ...",
							ContentLine2 = "From " + e.Device.DeviceName,
							NotificationKey = e.Device.InstanceID,
							BitmapIcon = caller_photo,
						};

						toast = new ToastPopUp(toastUIElement, true);

						toast.ToastActionTriggeredEvent += this.ToastActionHandler;
						toast.ToastClosedEvent += this.ToastClosed;

						toast.Show();
					}));

				StaticLogger.LogInfo("[###] Create phone call toast successfully");
			}
			catch (Exception ex)
			{
				StaticLogger.LogInfo("[###] An error happened when create phone call toast => " + ex.ToString());
			}

			return toast;
		}

		private void PhoneCallStateChangedHandler(object sender, ICStateChangeEventArgs e)
		{
			if (false == this._pairedDevices.ContainsKey(e.Device.InstanceID))
			{
				StaticLogger.LogInfo("[###] _pairedDevices list doesn't contain the device => " + e.Device.InstanceID);
				return;
			}

			switch (e.State)
			{
				case ICStateChangeEventArgs.LineState.AnswerOnPC:
					{
						// Suppose that only one device is allowed to answer phone call for each time
						if (!this._isAnsweringPhoneCall || this._currentActivatedDeviceID == e.Device.InstanceID)
						{
							this._pairedDevices[e.Device.InstanceID].phoneCallToast.RefreshToastUI("AnswerOnPC");
							StaticLogger.LogInfo("[###] Device, " + this._currentActivatedDeviceID.ToString() + " , is answering phone now.");
							this._currentActivatedDeviceID = e.Device.InstanceID;
							this._isAnsweringPhoneCall = true;
						}
						else
							StaticLogger.LogInfo("[###] There is a device, " + this._currentActivatedDeviceID.ToString() + " , answering phone.");

						break;
					}
				case ICStateChangeEventArgs.LineState.AnswerOnPhone:
					{
						// Suppose that only one device is allowed to answer phone call for each time
						if (!this._isAnsweringPhoneCall || this._currentActivatedDeviceID == e.Device.InstanceID)
						{
							this._pairedDevices[e.Device.InstanceID].phoneCallToast.RefreshToastUI("AnswerOnPC");
							StaticLogger.LogInfo("[###] Device, " + this._currentActivatedDeviceID.ToString() + " , is answering phone now.");
							this._currentActivatedDeviceID = e.Device.InstanceID;
							this._isAnsweringPhoneCall = true;
						}
						else
							StaticLogger.LogInfo("[###] There is a device, " + this._currentActivatedDeviceID.ToString() + " , answering phone.");

						break;
					}
				case ICStateChangeEventArgs.LineState.HangUpFromPC:
					{
						if (this._currentActivatedDeviceID == e.Device.InstanceID)
						{
							if (this._pairedDevices[e.Device.InstanceID].phoneCallToast != null)
								this._pairedDevices[e.Device.InstanceID].phoneCallToast.RefreshToastUI("HangUpFromPC");

							this._currentActivatedDeviceID = Guid.Empty;
							this._isAnsweringPhoneCall = false;
						}

						StaticLogger.LogInfo("[###] Device, " + e.Device.InstanceID + " , hangs up the phone call.");
						break;
					}
				case ICStateChangeEventArgs.LineState.HangUpFromPhone:
					{
						if (this._currentActivatedDeviceID == e.Device.InstanceID)
						{
							if (this._pairedDevices[e.Device.InstanceID].phoneCallToast != null)
								this._pairedDevices[e.Device.InstanceID].phoneCallToast.RefreshToastUI("HangUpFromPC");

							this._currentActivatedDeviceID = Guid.Empty;
							this._isAnsweringPhoneCall = false;
						}

						StaticLogger.LogInfo("[###] Device, " + e.Device.InstanceID + " , hangs up the phone call.");
						break;
					}
				case ICStateChangeEventArgs.LineState.UpdateTime:
					{
						break;
					}
				default:
					break;
			}
		}

		public void ToastActionHandler(object sender, ToastActionTriggeredEventArgs e)
		{
			if (this._pairedDevices.ContainsKey(e.PhoneCallToastID))
			{
				AirFusion.Manager.DeviceManager.AFUserDevice targetDevice = this._pairedDevices[e.PhoneCallToastID].phoneCallProperties.Device;

				switch (e.PhoneCallToastAction)
				{
					case PhoneCallToastActionType.Answer:
						{
							if (!this._isAnsweringPhoneCall || this._currentActivatedDeviceID == e.PhoneCallToastID)
							{
								try
								{
									this._incomingCallManager.Answer(targetDevice, null);
									this._currentActivatedDeviceID = targetDevice.InstanceID;
									this._isAnsweringPhoneCall = true;
								}
								catch (Exception ex)
								{
									StaticLogger.LogInfo("[###] An error happened when answer phone call => " + ex.ToString());
								}
							}
							else
								StaticLogger.LogInfo("[###] You cannot answer more than one phone call at each time.");

							break;
						}
					case PhoneCallToastActionType.Decline:
						{
							try
							{
								this._incomingCallManager.Decline(targetDevice, null);
							}
							catch (Exception ex)
							{
								StaticLogger.LogInfo("[###] An error happened when decline phone call => " + ex.ToString());
							}

							break;
						}
					case PhoneCallToastActionType.End:
						{
							try
							{
								this._incomingCallManager.HangUp(targetDevice, null);
							}
							catch (Exception ex)
							{
								StaticLogger.LogInfo("[###] An error happened when end phone call => " + ex.ToString());
							}

							break;
						}
					case PhoneCallToastActionType.Message:
						break;

					case PhoneCallToastActionType.Mute:
						break;

					case PhoneCallToastActionType.None:
						break;
				}
			}
		}

		public void ToastClosed(object sender, ToastClosedEventArgs e)
		{
			if (e.ToastPopUp.CurrentToast.NotificationKey == this._currentActivatedDeviceID)
			{
				this._currentActivatedDeviceID = Guid.Empty;
				this._isAnsweringPhoneCall = false;
			}

			FullyCloseToast(e.ToastPopUp);
		}

		private void FullyCloseToast(ToastPopUp toast)
		{
			PairedDevice pairedDevice = new PairedDevice()
			{
				isPhoneCallToastEnabled = this._pairedDevices[toast.CurrentToast.NotificationKey].isPhoneCallToastEnabled,
				phoneCallToast = null,
				phoneCallProperties = null,
			};
			this._pairedDevices.AddOrUpdate(toast.CurrentToast.NotificationKey, (key) => pairedDevice, (key, oldValue) => pairedDevice);

			toast.ToastActionTriggeredEvent -= this.ToastActionHandler;
			toast.Close();
			toast.ToastClosedEvent -= this.ToastClosed;
		}

		private BitmapSource ConvertBitmap(Bitmap source)
		{
			BitmapSource retImg = null;

			try
			{
				retImg = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
							source.GetHbitmap(),
							IntPtr.Zero,
							Int32Rect.Empty,
							BitmapSizeOptions.FromEmptyOptions());
			}
			catch (Exception ex)
			{
				string imgSource = @"pack://application:,,,/AirFusionCustomToast;component/Resources/Caller.png";
				retImg = new BitmapImage(new Uri(imgSource, UriKind.Absolute));
			}

			return retImg;
		}
	}
}