﻿using AirFusion.Common;
using AirFusion.Common.Network;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
// 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
/*
using InTheHand.Net;
using InTheHand.Net.Sockets;
*/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class UserDeviceManager
	{
		#region Misellaneuous

		/// <summary>
		/// Initializes the local device.
		/// </summary>
		/// <exception cref="System.Exception" />
		private void InitializeLocalDevice()
		{
			try
			{
				StaticLogger.TrackEnter();

				// read information of current device
				try
				{
					this.ReadLocalDeviceFixedInformation();
				}
				catch (Exception ex)
				{
					throw new Exception("Failed reading fixed information of current device.", ex);
				}

				// read configuration
				try
				{
					this.ReadConfiguration();
				}
				catch (Exception ex)
				{
					throw new Exception("Failed reading configuration.", ex);
				}

				// save the instance ID if it's not present yet
				if (!this.Configuration.LocalInstanceID.HasValue || this.Configuration.LocalInstanceID.Value.CompareTo(Guid.Empty) == 0)
				{
					this.Configuration.LocalInstanceID = this.LocalDevice.InstanceID;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Initializes the network observer.
		/// </summary>
		private void InitializeNetworkObserver()
		{
			try
			{
				StaticLogger.TrackEnter();

				// create a network observer
				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNNetworkObserver)))
				{
					this._networkObserver = new NetworkObserver();
					this._networkObserver.NetworkChanged += (s, e) =>
					{
						try
						{
							StaticLogger.LogTrace("NetworkChanged event raised.");

							if (e.NetworkList != null)
							{
								List<InstanceProfile.NetworkInfo> list = new List<InstanceProfile.NetworkInfo>();
								foreach (var networkInfo in e.NetworkList)
								{
									try
									{
										list.Add(networkInfo.Transform());
									}
									catch (Exception ex)
									{
										StaticLogger.LogWarn("Failed converting network information, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}
								}

								this.LocalProfile.Network = list.ToArray();

								lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNDandelionAdapter)))
								{
									// update profile on Dandelion
									if (this._dandelionAdapter != null)
									{
										StaticLogger.LogTrace("Set profile through Dandelion.");

										this._dandelionAdapter.SetProfile(this.LocalProfile);
									}
								}
							}
						}
						catch (Exception ex)
						{
							StaticLogger.LogError("Failed updating profile, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						}
					};

					// create a network observable
					this._networkObservable = new NetworkObservable();
					this._networkObservable.Subscribe(this._networkObserver);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Initializes the multicast DNS adapter.
		/// </summary>
		private void InitializeMulticastDnsAdapter()
		{
			try
			{
				StaticLogger.TrackEnter();

				// TODO: the interface of MulticastDnsAdapter should be same as DandelionAdapter

				Debug.Assert(this._mdnsAdapter != null);

				// initialize mDNS adapter
				StaticLogger.LogTrace("Initialize mDNS service...");
				this._mdnsAdapter.Initialize();

				// register mDNS service
				StaticLogger.LogTrace("Register mDNS service...");
				Debug.Assert(this.LocalDevice != null);
				this._mdnsAdapter.RegisterService(this.LocalDevice.ToInstance());

				// start observation
				StaticLogger.LogTrace("Start mDNS observation...");
				this._mdnsAdapter.AddressUpdated += this.OnMulticastDnsAdapterAddressUpdated;
				this._mdnsAdapter.ServiceFound += this.OnMulticastDnsAdapterServiceFound;
				this._mdnsAdapter.ServiceLost += this.OnMulticastDnsAdapterServiceLost;
				this._mdnsAdapter.StartObservation();
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed initializing mDNS adapter, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				this._mdnsAdapter = null;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Initializes the dandelion adapter.
		/// </summary>
		private void InitializeDandelionAdapter()
		{
			try
			{
				StaticLogger.TrackEnter();

				// TODO: the interface of MulticastDnsAdapter should be same as DandelionAdapter

				if (this.Configuration != null && !string.IsNullOrEmpty(this.Configuration.DandelionServerHost))
				{
					if (this.Configuration.DandelionServerSecure.HasValue)
					{
						this._dandelionAdapter = new DandelionAdapter(
							this.Configuration.DandelionServerHost,
							this.Configuration.DandelionServerPort,
							this.Configuration.DandelionServerSecure.Value);
					}
					else
					{
						this._dandelionAdapter = new DandelionAdapter(this.Configuration.DandelionServerHost, this.Configuration.DandelionServerPort);
					}
				}
				else
				{
					this._dandelionAdapter = new DandelionAdapter();
				}

				Debug.Assert(this.LocalProfile != null);
				Debug.Assert(this.LocalProfile.InstanceID.HasValue);
				this._dandelionAdapter.Initialize(this.LocalProfile.InstanceID.Value);
				this._dandelionAdapter.SetProfile(this.LocalProfile);
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed preparing Dandelion, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				this._dandelionAdapter = null;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
		/*
		/// <summary>
		/// Initializes the Bluetooth.
		/// </summary>
		private void InitializeBluetooth()
		{
			try
			{
				StaticLogger.TrackEnter();

				if (CommonUtils.IsBluetoothEnabled())
				{
					// start Bluetooth listener
					try
					{
						StaticLogger.LogTrace("Starting Bluetooth listener...");
						this.StartBluetoothListener();
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("Failed starting Bluetooth listener, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
				}
				else
				{
					StaticLogger.LogWarn("There's no Bluetooth installed or not enabled.");
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
		*/

		/// <summary>
		/// Initializes the web service.
		/// </summary>
		/// <param name="manager">The manager.</param>
		/// <exception cref="System.ArgumentNullException" />
		private void InitializeWebServices(WebServicesManager manager)
		{
			try
			{
				StaticLogger.TrackEnter();

				if (manager == null)
				{
					throw new ArgumentNullException("manager");
				}

				this.WebServiceManager = manager;

				// device information service
				{
					StaticLogger.LogTrace("Initializing device information web service...");
					this.DeviceInformationService = new DeviceInformationService(this.WebServiceManager);
					this.DeviceInformationService.RegisterRequestCallback(this.OnDeviceInformationRequested);
					this.DeviceInformationService.RegisterUpdateCallback(this.OnDeviceInformationUpdated);
				}

				// pairing service
				{
					StaticLogger.LogTrace("Initializing pairing web service...");
					this.PairingService = new PairingService(this.WebServiceManager);
					this.PairingService.RegisterPairingInvitationCallback(this.OnPairingInvited);
					this.PairingService.RegisterPairingRequestCallback(this.OnPairingRequested);
					this.PairingService.RegisterPairingResponseCallback(this.OnPairingResponsed);
					this.PairingService.RegisterPairingStatusRequestCallback(this.OnPairingStatusRequested);
					this.PairingService.RegisterPairingStatusUpdateCallback(this.OnPairingStatusUpdated);
				}

				// connection helper service
				{
					StaticLogger.LogTrace("Initializing connection helper web service...");
					this.ConnectionHelperService = new ConnectionHelperService(this.WebServiceManager);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads the fixed information of local device.
		/// </summary>
		private void ReadLocalDeviceFixedInformation()
		{
			try
			{
				StaticLogger.TrackEnter();

				this._localDevice = new AFUserDevice();
				this._localDevice.Address = IPAddress.Loopback; // TODO: Here I do not know which address is preferred by other device, it should be a collection but I am using loopback for myself
				this._localDevice.Owner = this;
				this._localDevice._pairStatus = AFUserDevice.PairingStatus.UnPaired;

				// must
				this._localDevice.InstanceID = this.Configuration != null && this.Configuration.LocalInstanceID.HasValue && this.Configuration.LocalInstanceID.Value.CompareTo(Guid.Empty) != 0 ? this.Configuration.LocalInstanceID.Value : Guid.NewGuid();
				this._localDevice.OSType = DefaultOSType;
				this._localDevice.Services = new AFUserDevice.ServicesInfo();

				// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
				/*
				this._localDevice.Bluetooth = new AFUserDevice.BluetoothInfo()
				{
					Address = CommonUtils.IsBluetoothEnabled() ? CommonUtils.GetBluetoothAddress() : null
				};
				*/

				// optional
				{
					// display
					{
						// TODO: to change the property when display change
						// TODO: implement this using abstraction
						double width, height;
						AFSystemInformation.GetPrimaryScreenResolution(out width, out height);

						this._localDevice.Display = new AFUserDevice.DisplayInfo()
						{
							Rotation = null, // the property is not used currently
							Width = (int)width,
							Height = (int)height,
							Navigation = false // Windows device never get a navigation bar
						};
					}

					try
					{
						// read WMI values for some properties
						AirFusion.Common.Wmi.Win32.WmiComputerSystem wmiWin32ComputerSystem = AirFusion.Common.Wmi.WmiReader<AirFusion.Common.Wmi.Win32.WmiComputerSystem>.Get().FirstOrDefault();
						AirFusion.Common.Wmi.Win32.WmiOperatingSystem wmiWin32OperatingSystem = AirFusion.Common.Wmi.WmiReader<AirFusion.Common.Wmi.Win32.WmiOperatingSystem>.Get().FirstOrDefault();

						this._localDevice.ModelName = wmiWin32ComputerSystem != null ? wmiWin32ComputerSystem.Model : null;
						this._localDevice.Manufacturer = wmiWin32ComputerSystem != null ? wmiWin32ComputerSystem.Manufacturer : null;
						this._localDevice.OS = wmiWin32OperatingSystem != null ? wmiWin32OperatingSystem.Caption + wmiWin32OperatingSystem.CSDVersion : null;
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed reading WMI values, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}

					try
					{
						this._localDevice.DeviceName = Environment.MachineName;
						this._localDevice.OSVersion = Environment.OSVersion.Version.ToString();
						this._localDevice.UserName = Environment.UserName;
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed reading Environment values, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}

					try
					{
						this._localDevice.ApplicationVersion = Assembly.GetEntryAssembly().GetName().Version.ToString();
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed reading Assembly values, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads configuration.
		/// </summary>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.Security.SecurityException" />
		/// <exception cref="System.UnauthorizedAccessException" />
		private void ReadConfiguration()
		{
			try
			{
				StaticLogger.TrackEnter();

				if (this.Configuration == null)
				{
					this.Configuration = new DeviceConfiguration(RegDeviceManagementHive, RegDeviceManagementKey);
				}

				Debug.Assert(this.Configuration != null);
				lock (CommonUtils.GetLocker(this.Configuration))
				{
					// local instance ID
					if (this.Configuration.LocalInstanceID.HasValue && this.Configuration.LocalInstanceID.Value.CompareTo(Guid.Empty) != 0)
					{
						this.LocalDevice.InstanceID = this.Configuration.LocalInstanceID.Value;
					}
					else
					{
						this.Configuration.LocalInstanceID = this.LocalDevice.InstanceID;
					}

					// 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
					/*
					// local Bluetooth address
					if (this.Configuration.LocalBluetoothAddress != null && !this.Configuration.LocalBluetoothAddress.Equals(BluetoothAddress.None))
					{
						this._localDevice.Bluetooth = new AFUserDevice.BluetoothInfo()
						{
							Address = this.Configuration.LocalBluetoothAddress
						};
					}
					*/

					// read paired devices
					Debug.Assert(this.DeviceStore != null);
					lock (CommonUtils.GetLocker(new Tuple<object, object>(this, this.DeviceStore)))
					{
						if (this.Configuration.PairedDeviceList != null)
						{
							foreach (var pairedDevice in this.Configuration.PairedDeviceList)
							{
								if (pairedDevice == null)
								{
									continue;
								}

								var device = this.DeviceStore.AddOrUpdate(pairedDevice);
								device._pairStatus = AFUserDevice.PairingStatus.Paired; // assign to _pairStatus rather than PairStatus to prevent Paired event being invoked
							}
						}
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Misellaneuous

		#region Bluetooth

		// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
		/*
		/// <summary>
		/// Starts the Bluetooth listener if it's not listening.
		/// </summary>
		/// <exception cref="System.NotSupportedException" />
		private void StartBluetoothListener()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prequisite
				if (!CommonUtils.IsBluetoothEnabled())
				{
					throw new NotSupportedException("There's no Bluetooth installed or not enabled.");
				}

				if (this._bluetoothListener == null)
				{
					this._bluetoothListener = new BluetoothListener(BluetoothServiceUuid);
					this._bluetoothListener.Start();
				}

				StaticLogger.LogTrace("The Bluetooth listener has been started.");

				this._bluetoothListener.BeginAcceptBluetoothClient(delegate(IAsyncResult result)
				{
					if (result != null && result.IsCompleted)
					{
						BluetoothListener listener = result.AsyncState as BluetoothListener;
						if (listener != null)
						{
							try
							{
								BluetoothClient client = listener.EndAcceptBluetoothClient(result);
								if (client != null && client.RemoteEndPoint != null && client.RemoteEndPoint.Address != null)
								{
									StaticLogger.LogInfo("A Bluetooth client '" + client.RemoteEndPoint.Address.GetStringFormat() + "' has connected.");
								}
							}
							catch (Exception ex)
							{
								StaticLogger.LogWarn("Failed accepting Bluetooth client, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						}

						// TBD:
						// hold the socket connection
					}
				}, this._bluetoothListener);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops the Bluetooth listener if it's listening.
		/// </summary>
		private void StopBluetoothListener()
		{
			try
			{
				StaticLogger.TrackEnter();

				if (this._bluetoothListener != null)
				{
					lock (this._bluetoothListener)
					{
						this._bluetoothListener.Stop();
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
		*/

		#endregion Bluetooth

		#region MulticastDnsAdapter

		/// <summary>
		/// Starts advertising through mDNS.
		/// </summary>
		private void StartMulticastDnsAdvertise()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// TODO
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops advertising through mDNS.
		/// </summary>
		private void StopMulticastDnsAdvertise()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// TODO
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts discovery through mDNS.
		/// </summary>
		private void StartMulticastDnsDiscovery(double duration = 0.0)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// return if the DiscoveryEvent is not set
				if (this.DiscoveryEvent == null)
				{
					return;
				}

				// for each instance in instance store, fire DiscoveryEvent event
				foreach (var instance in this._mdnsAdapter.GetInstances().ToArray())
				{
					if (instance.InstanceID.HasValue)
					{
						// get from device store if such device already exist, or create one if not
						var device = this.DeviceStore.GetDevice(instance.InstanceID.Value);

						if (device != null)
						{
							// fire DiscoveryEvent
							if (this.IsDiscovering && this.DiscoveryEvent != null)
							{
								CommonUtils.InvokeEvent<DiscoveryEventArgs>(
									new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent),
									this,
									new DiscoveryEventArgs(device, DiscoveryEventArgs.DiscoveryEventType.DeviceFoundOrUpdate));
							}
						}
						else
						{
							StaticLogger.LogWarn("The device '" + instance.InstanceID + "' is not found in device store, skipped.");
						}
					}
					else
					{
						StaticLogger.LogWarn("The instance ID of instance '" + instance.InstanceName + "' is unspecified and the instance is unconnectable, skipped.");
					}
				}

				try
				{
					// trigger an immediate mDNS browse
					this._mdnsAdapter.ImmediateBrowse(duration > 0 ? duration : 5000);
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed triggering an immediate browse, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed starting mDNS discovery, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops discovery through mDNS.
		/// </summary>
		private void StopMulticastDnsDiscovery()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// TODO
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Queries the specified device for information through local web service asynchronously.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.Exception" />
		private async Task<AFUserDevice> QueryWebServiceForInformationAsync(MulticastDnsAdapter.MDnsInstance instance)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameter
				StaticLogger.LogDebug("instance: " + (instance != null ? "\r\n" + instance.ToString() : "N/A"));

				// check parameter
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}
				if (!instance.IsConnectable())
				{
					throw new ArgumentException("The address or the port number of instance '" + instance.InstanceName + "' is not specified.");
				}

				// check prerequisite
				if (this.DeviceInformationService == null)
				{
					throw new InvalidOperationException("The device information service is not initialized yet.");
				}

				var response = await this.DeviceInformationService.RequestAsync(instance.Address, instance.Port.Value).ConfigureAwait(continueOnCapturedContext: false);
				if (response != null)
				{
					if (response.ResponseData.InstanceID.HasValue)
					{
						// get from device store if such device already exist, or create one if not
						var device = this.DeviceStore.GetDevice(response.ResponseData.InstanceID.Value) ?? new AFUserDevice();
						lock (CommonUtils.GetLocker(new Tuple<object, object>(this, device)))
						{
							device.FillFromIfSpecified(response.ResponseData);
						}

						// update the configuration if it's Paired
						if (device.PairStatus == AFUserDevice.PairingStatus.Paired)
						{
							Debug.Assert(this.Configuration != null);
							this.Configuration.StoreDevice(device);
						}

						return device;
					}
					else
					{
						throw new Exception("The instance ID of the response is unspecified.");
					}
				}
				else
				{
					throw new Exception("The response of the web service is invalid.");
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Determines if the specified instance ID is found through mDNS.
		/// </summary>
		/// <param name="instanceID">The instance ID</param>
		private bool IsFoundByMulticastDns(Guid instanceID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return false;
				}

				return this._mdnsAdapter.ExistInstance(instanceID);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion MulticastDnsAdapter

		#region DandelionAdapter

		/// <summary>
		/// Starts advertising through Dandelion.
		/// </summary>
		private void StartDandelionAdvertise()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._dandelionAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the Dandelion adpater is not initialized.");
					return;
				}

				// begin Dandelion advertisement
				this._dandelionAdapter.BeginAdvertise();
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed starting Dandelion advertisement, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops advertising through Dandelion.
		/// </summary>
		private void StopDandelionAdvertise()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._dandelionAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the Dandelion adpater is not initialized.");
					return;
				}

				// end Dandelion advertisement
				this._dandelionAdapter.EndAdvertise();
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed to stop Dandelion advertisement, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts discovery through Dandelion.
		/// </summary>
		private void StartDandelionDiscovery()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._dandelionAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the Dandelion adpater is not initialized.");
					return;
				}

				this._dandelionAdapter.BeginGetNearbyDevices(this.OnDandelionAdapterNearbyScanCallback);
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed starting Dandelion discovery, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops discovery through Dandelion.
		/// </summary>
		private void StopDandelionDiscovery()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this._dandelionAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the Dandelion adpater is not initialized.");
					return;
				}

				// TODO:
				// implement this if Dandelion has a method to stop discovery
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed to stop Dandelion discovery, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion DandelionAdapter

		#region AFUserDevice

		/// <summary>
		/// Gets the IP address of the specified device through mDNS address reconfirmation &amp; Dandelion asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <param name="returnIfAnyConnectableAddress">Whether to return if there's any connectable address rather than testing all returned addresses.</param>
		/// <returns>The result of the operation.</returns>
		/// <exception cref="System.ArgumentNullException" />
		internal async Task<ReconfirmAddressResult> ReconfirmAddressAsync(AFUserDevice device, bool returnIfAnyConnectableAddress = true)
		{
			try
			{
				StaticLogger.TrackEnter();

				var taskList = new List<Task<IPEndPoint[]>>();

				// check parameters
				if (device == null)
				{
					throw new ArgumentNullException("device");
				}

				// log parameters
				StaticLogger.LogDebug("device: " + (device != null ? device.ToString() : "N/A"));

				// mDNS
				try
				{
					if (this._mdnsAdapter != null)
					{
						var task = this._mdnsAdapter.ReconfirmEndpointAsync(device.InstanceID);
						StaticLogger.LogInfo("A task(#" + task.Id + ") has been created to reconfirm the address through mDNS.");
						taskList.Add(task);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed starting reconfirming endpoint through mDNS, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}

				// Dandelion
				try
				{
					if (this._dandelionAdapter != null)
					{
						var task = this._dandelionAdapter.GetEndpointAsync(device.InstanceID);
						StaticLogger.LogInfo("A task(#" + task.Id + ") has been created to reconfirm the address through Dandelion.");
						taskList.Add(task);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed starting getting endpoint through Dandelion, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}

				// wait until an approach success
				var result = new ReconfirmAddressResult();
				while (taskList != null && taskList.Count > 0)
				{
					var task = await Task.WhenAny(taskList);
					if (!task.IsFaulted && task.Result != null)
					{
						StaticLogger.LogInfo("The returned endpoint are " + string.Join<IPEndPoint>(";", task.Result) + ".");

						if (this.ConnectionHelperService != null)
						{
							foreach (var endPoint in task.Result)
							{
								if (endPoint == null || endPoint.Address == null || endPoint.Port == 0)
								{
									continue;
								}

								try
								{
									// validate the endpoint by pinging if the connection helper service has been created
									StaticLogger.LogInfo("Try to ping " + (endPoint.Address != null ? endPoint.Address.ToString() : "N/A") + ":" + endPoint.Port + " to ensure the endpoint is valid.");

									var response = await this.ConnectionHelperService.PingAsync(endPoint.Address, (ushort)endPoint.Port).ConfigureAwait(continueOnCapturedContext: false);
									result.EndPointConnectionStatus[endPoint] = ConnectionStatus.Connectable;
									if (response.IsMessageResponse)
									{
										StaticLogger.LogInfo("ConnectionHelperService.PingAsync() is done successfully: " + response.Message);
									}
									else
									{
										StaticLogger.LogInfo("ConnectionHelperService.PingAsync() is done successfully.");
									}

									// break if returnIfAnyConnectableAddress=true
									if (returnIfAnyConnectableAddress)
									{
										break;
									}
								}
								catch (ArgumentException ex)
								{
									Trace.Fail("The exception is an ArgumentException: " + ex.Message);
									throw;
								}
								catch (TimeoutException)
								{
									StaticLogger.LogWarn("Failed pinging to '" + device + "' due to timeout, considering as unconnectable.");
									result.EndPointConnectionStatus[endPoint] = ConnectionStatus.Unconnectable;
								}
								catch (WebServiceException ex)
								{
									StaticLogger.LogWarn("ConnectionHelperService.PingAsync() returns: " + ex.Message);
									result.EndPointConnectionStatus[endPoint] = ConnectionStatus.Connectable; // though the status code is not correct, but it is considered as connectable
								}
								catch (Exception ex)
								{
									StaticLogger.LogWarn("Failed pinging device '" + device + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									result.EndPointConnectionStatus[endPoint] = ConnectionStatus.Unconnectable;
								}
							}
						}
						else
						{
							StaticLogger.LogInfo("Since we don't know which is connectable, just add them all.");

							foreach (var endPoint in task.Result)
							{
								if (endPoint == null || endPoint.Address == null || endPoint.Port == 0)
								{
									continue;
								}

								if (result.EndPointConnectionStatus.ContainsKey(endPoint) == false)
								{
									result.EndPointConnectionStatus.Add(endPoint, ConnectionStatus.Unknown);
								}
							}
						}
					}
					else
					{
						StaticLogger.LogError("Failed getting endpoint from task#" + task.Id + (task.Exception != null ? ", exception:\r\n" + CommonUtils.GetExceptionMessages(task.Exception) : "."));
					}

					if (task.IsCompleted)
					{
						taskList.Remove(task);
					}
				}

				// choose the best endpoint
				{
					var firstConnectableEndpoint = result.EndPointConnectionStatus.Where(item => item.Value == ConnectionStatus.Connectable).Select(item => item.Key).FirstOrDefault();
					if (firstConnectableEndpoint == null)
					{
						firstConnectableEndpoint = result.EndPointConnectionStatus.Where(item => item.Value == ConnectionStatus.Unknown).Select(item => item.Key).FirstOrDefault();
					}

					if (firstConnectableEndpoint != null)
					{
						// update the address and port
						device.Address = firstConnectableEndpoint.Address;
						device.WebServicePort = (ushort)firstConnectableEndpoint.Port;

						// TODO: store the device using RegistryHolders
						if (device.PairStatus == AFUserDevice.PairingStatus.Paired)
						{
							this.Configuration.StoreDevice(device);
						}
					}
					else
					{
						StaticLogger.LogInfo("There's or connectable address found.");
					}
				}

				return result;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends out the pairing invitation to the specified device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		internal async Task PairingInviteAsync(AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("device: " + (device != null ? device.ToString() : "N/A"));

				// check parameters
				if (device == null)
				{
					throw new ArgumentNullException("device");
				}

				// check prerequisite
				if (device.PairStatus == AFUserDevice.PairingStatus.Paired)
				{
					throw new InvalidOperationException("The device is already paired.");
				}
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address of device '" + device + "' is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.PairingService != null, "The 'PairingService' is not specified.");
				Trace.Assert(this.DeviceStore != null, "The 'DeviceStore' is not specified.");
				Trace.Assert(this.LocalDevice != null, "The 'LocalDevice' is not specified.");
				Trace.Assert(device.InstanceID.Equals(Guid.Empty) == false, "The 'device.InstanceID' is Guid.Empty.");

				// add the device into the device store if it's not present yet
				this.DeviceStore.AddOrUpdate(device);

				// check if the instance exist in mDNS instance store
				if (this._mdnsAdapter != null && this._mdnsAdapter.GetInstance(device.InstanceID) != null)
				{
					StaticLogger.LogTrace("Send pairing invitation through local web service since the instance is found by mDNS.");

					var pairingInvitation = new PairingInvitationAdapter.PairingInvitationInfo()
					{
						InstanceID = this.LocalDevice.InstanceID,
						DeviceName = this.LocalDevice.DeviceName,
						ModelName = this.LocalDevice.ModelName,
						UserName = this.LocalDevice.UserName,
						CommunicationPort = this.LocalDevice.WebServicePort,
						SessionKey = null, // TBD
					};

					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Send pairing invitation to '" + address + ":" + port + "'...");

					try
					{
						var response = await this.PairingService.InviteAsync(address, port, pairingInvitation).ConfigureAwait(continueOnCapturedContext: false);
						if (response.IsMessageResponse)
						{
							StaticLogger.LogInfo("PairingService.InviteAsync() is done successfully: " + response.Message);
						}
						else
						{
							StaticLogger.LogInfo("PairingService.InviteAsync() is done successfully.");
						}
					}
					catch (ArgumentException ex)
					{
						Trace.Fail("The exception is an ArgumentException: " + ex.Message);
						throw;
					}
					catch (WebServiceException ex)
					{
						StaticLogger.LogError("PairingService.InviteAsync() returns: " + ex.Message);
						throw;
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed sending pairing invitation to '" + device + "' through web service: " + ex.Message);
						throw;
					}
				}
				else if (this._dandelionAdapter != null)
				{
					StaticLogger.LogTrace("Send pairing invitation through Dandelion since the instance is not found by mDNS.");

					try
					{
						await this._dandelionAdapter.SendPairingInvitationAsync(device.InstanceID).ConfigureAwait(continueOnCapturedContext: false);
					}
					catch (ArgumentException ex)
					{
						Trace.Fail("The exception is an ArgumentException: " + ex.Message);
						throw;
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed sending pairing invitation to '" + device + "' through Dandelion: " + ex.Message);
						throw;
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion AFUserDevice
	}
}
