﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	/// <summary>
	/// Represents the session ID.
	/// </summary>
	public sealed class PairingSessionId : ISessionId
	{
		#region Private/Protected/Internal Variables

		private byte[] _sessionId;

		#endregion Private/Protected/Internal Variables

		#region Override ISessionId

		/// <summary>
		/// Gets the byte array format of session ID.
		/// </summary>
		/// <returns></returns>
		/// <exception cref="System.InvalidOperationException">This instance hasn't been initialized.</exception>
		public byte[] GetByteArray()
		{
			if (this._sessionId == null)
			{
				throw new InvalidOperationException("This instance hasn't been initialized.");
			}

			return this._sessionId;
		}

		/// <summary>
		/// Sets the byte array format of session ID.
		/// </summary>
		/// <param name="byteArray">The byte array.</param>
		public void SetByteArray(byte[] byteArray)
		{
			if (byteArray == null)
			{
				throw new ArgumentNullException("byteArray");
			}
			if (byteArray.Length == 0)
			{
				throw new ArgumentException("The length of the byte array is zero.");
			}

			this._sessionId = byteArray;
		}

		#endregion Override ISessionId

		#region Override Object

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			return this._sessionId != null ? this._sessionId.ToHexString() : "N/A";
		}

		#endregion Override Object

		#region Private/Protected/Internal Methods

		/// <summary>
		/// Prevents a default instance of the <see cref="PairingSessionId"/> class from being created.
		/// </summary>
		private PairingSessionId()
		{
			this._sessionId = null;
		}

		#endregion Private/Protected/Internal Methods

		#region Public Static Methods

		/// <summary>
		/// Creates the session ID with the specified length.
		/// </summary>
		/// <param name="length">The length.</param>
		/// <returns></returns>
		public static ISessionId CreateSessionId(uint length = AFUserDevice.DefaultSessionIdLength)
		{
			if (length == 0)
			{
				throw new ArgumentException("The length is zero.");
			}

			PairingSessionId sessionId = new PairingSessionId();
			sessionId._sessionId = new byte[length];
			new Random().NextBytes(sessionId._sessionId);

			return sessionId;
		}

		/// <summary>
		/// Creates the session ID with the specified byte array.
		/// </summary>
		/// <param name="byteArray">The byte array.</param>
		/// <returns></returns>
		public static ISessionId CreateSessionId(byte[] byteArray)
		{
			PairingSessionId sessionId = new PairingSessionId();
			sessionId.SetByteArray(byteArray);

			return sessionId;
		}

		#endregion Public Static Methods
	}
}