﻿using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of wakeup (POST /paging).
	/// </summary>
	internal sealed class WakeUpAdapter : WebServiceAdapter<DBNull, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/paging";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="WakeUpAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public WakeUpAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
			this.Timeout = TimeSpan.FromMilliseconds(10000);
		}

		/// <summary>
		/// Sends a wakeup request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> WakeUpAsync(IPAddress address, ushort port)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri);

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
