﻿using AirFusion.Common;
using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of device information (request) (GET /information/{InstanceID}).
	/// </summary>
	internal sealed class DeviceInformationRequestAdapter : WebServiceAdapter<DBNull, DeviceInformation>
	{
		#region Definitions

		internal const string WebServiceUri = "/information";
		internal static HttpMethod WebServiceMethod = HttpMethod.Get;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="DeviceInformationRequestAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public DeviceInformationRequestAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Requests the device information from the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="instanceID">The instance ID of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestAsync(IPAddress address, ushort port, Guid? instanceID = null)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + (instanceID.HasValue ? "/" + instanceID.Value.GetStringFormat() : string.Empty)).Uri);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
