﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;

namespace AirFusion.Manager.FeatureHelper
{
	/// <summary>
	/// Represents the configuration helper.
	/// </summary>
	public sealed partial class ConfigurationHelper
	{
		#region Statics Properties/Variables/Methods

		internal static IDictionary<UserDeviceManager, ConfigurationHelper> DeviceManagerMapping = new Dictionary<UserDeviceManager, ConfigurationHelper>();

		#endregion Statics Properties/Variables/Methods

		#region Properties

		/// <summary>
		/// Gets the configuration service.
		/// </summary>
		internal ConfigurationService ConfigurationService { get; private set; }

		#endregion Properties

		#region Private/Protected/Internal Variables

		private UserDeviceManager _deviceManager;

		#endregion Private/Protected/Internal Variables

		#region Events

		/// <summary>
		/// Occurs when display configuration of remote device is updated.
		/// </summary>
		public event DisplayConfigurationUpdatedEventHandler DisplayConfigurationUpdated;

		#endregion Events

		#region Public Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="ConfigurationHelper"/> class.
		/// </summary>
		internal ConfigurationHelper(UserDeviceManager deviceManager, WebServicesManager webManager)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (deviceManager == null)
				{
					throw new ArgumentNullException("deviceManager");
				}
				if (webManager == null)
				{
					throw new ArgumentNullException("webManager");
				}

				// debug assertion
				Debug.Assert(DeviceManagerMapping != null);

				// check prerequisite
				if (DeviceManagerMapping.ContainsKey(deviceManager))
				{
					throw new InvalidOperationException("The specified device manager already have a configuration helper.");
				}

				this._deviceManager = deviceManager;
				DeviceManagerMapping.Add(deviceManager, this);

				// configuration service
				{
					this.ConfigurationService = new ConfigurationService(webManager);
					this.ConfigurationService.RegisterDisplayUpdateRequestCallback(this.OnDisplayConfigurationUpdated);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods

		#region Callbacks

		/// <summary>
		/// Called when the display configuration of a device is changed.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="targetInstanceID">The instance ID of device to be updated.</param>
		/// <param name="displayConfiguration">The display configuration.</param>
		/// <returns>The response to the request.</returns>
		private DisplayConfigurationUpdateAdapter.WebServiceResponse OnDisplayConfigurationUpdated(IPAddress clientIP, Guid targetInstanceID, DisplayConfigurationUpdateAdapter.DisplayConfiguration displayConfiguration)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("targetInstanceID: " + targetInstanceID.GetStringFormat());
				StaticLogger.LogDebug("displayConfiguration: " + (displayConfiguration != null ? displayConfiguration.ToString() : "N/A"));

				// debug assertion
				Debug.Assert(this._deviceManager != null);
				Debug.Assert(this._deviceManager.DeviceStore != null);

				// check prerequisite
				if (displayConfiguration == null)
				{
					StaticLogger.LogError("The display configuration is unspecified.");
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The display configuration is unspecified.");
				}
				if (targetInstanceID.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogError("The instance ID '" + targetInstanceID.GetStringFormat() + "' is invalid.");
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + targetInstanceID.GetStringFormat() + "' is invalid.");
				}
				if (targetInstanceID.CompareTo(this._deviceManager.LocalDevice.InstanceID) == 0)
				{
					StaticLogger.LogError("The display configuration of local device can't be updated by remote device.");
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.Unauthorized, "The display configuration of local device can't be updated by remote device.");
				}

				// get device from device list
				AFUserDevice device = this._deviceManager.DeviceStore.GetDevice(targetInstanceID);
				if (device == null)
				{
					StaticLogger.LogWarn("The device '" + targetInstanceID.GetStringFormat() + "' is unrecognized.");
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotFound, "The device '" + targetInstanceID.GetStringFormat() + "' is unrecognized. Please make sure you have both devices paired previously.");
				}

				// debug assertion
				Debug.Assert(displayConfiguration != null);
				Debug.Assert(device != null);

				// fill/update values if such values are specified
				device.Display.FillFromIfSpecified(displayConfiguration);

				// fire DisplayConfigurationUpdated event
				if (this.DisplayConfigurationUpdated != null)
				{
					// TODO:
					// fire event only when updated

					StaticLogger.LogTrace("Fire DisplayConfigurationUpdated event...");
					CommonUtils.InvokeEvent<DisplayConfigurationUpdatedEventArgs>(new EventHandler<DisplayConfigurationUpdatedEventArgs>(this.DisplayConfigurationUpdated), this, new DisplayConfigurationUpdatedEventArgs(device, device.Display));

					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
				else
				{
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
