﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Diagnostics;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service manager of configuration (/configuration/*).
	/// </summary>
	internal sealed class ConfigurationService
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle display configuration update request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="targetInstanceID">The instance ID of device to be updated.</param>
		/// <param name="displayConfiguration">The updated device information.</param>
		public delegate DisplayConfigurationUpdateAdapter.WebServiceResponse DisplayConfigurationUpdateCallback(IPAddress clientIP, Guid targetInstanceID, DisplayConfigurationUpdateAdapter.DisplayConfiguration displayConfiguration);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		private DisplayConfigurationUpdateAdapter _displayUpdateAdapter;
		private KeyboardConfigurationUpdateAdapter _keyboardUpdateAdapter;
		private DisplayConfigurationUpdateCallback _displayConfigurationUpdateCallback;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="ConfigurationService"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public ConfigurationService(WebServicesManager webServiceManager)
		{
			this._displayUpdateAdapter = new DisplayConfigurationUpdateAdapter(webServiceManager);
			this._keyboardUpdateAdapter = new KeyboardConfigurationUpdateAdapter(webServiceManager);

			this._displayUpdateAdapter.RegisterRequestCallback(this.OnDisplayAdapterUpdateRequestReceived);

			this._displayConfigurationUpdateCallback = null;
		}

		/// <summary>
		/// Registers the callback function of display configuration update request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterDisplayUpdateRequestCallback(DisplayConfigurationUpdateCallback callback)
		{
			this._displayConfigurationUpdateCallback = callback;
		}

		/// <summary>
		/// Sends a rotate request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="rotation">The rotation degree.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<DisplayConfigurationUpdateAdapter.WebServiceResponse> RotateAsync(IPAddress address, ushort port, DisplayRotation rotation)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._displayUpdateAdapter != null);
				return await this._displayUpdateAdapter.RotateAsync(address, port, rotation).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Switches the keyboard mode on the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="mode">The keyboard mode.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<KeyboardConfigurationUpdateAdapter.WebServiceResponse> SwitchKeyboardModeAsync(IPAddress address, ushort port, KeyboardConfigurationUpdateAdapter.KeyboardMode mode)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._displayUpdateAdapter != null);
				return await this._keyboardUpdateAdapter.SetKeyboardModeAsync(address, port, mode).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods

		#region Callbacks

		/// <summary>
		/// Called when a display configuration update request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private DisplayConfigurationUpdateAdapter.WebServiceResponse OnDisplayAdapterUpdateRequestReceived(IPAddress address, ushort port, DisplayConfigurationUpdateAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._displayConfigurationUpdateCallback == null)
				{
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered callback function.");
				}
				if (remoteRequest == null)
				{
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// get instance ID
				Guid instanceID;
				{
					string webServiceUriTrimSlashes = DisplayConfigurationUpdateAdapter.WebServiceUri.TrimStart('/');
					string requestUriTrimSlashes = remoteRequest.RequestUri.ToString().TrimStart('/');

					if (!requestUriTrimSlashes.StartsWith(webServiceUriTrimSlashes))
					{
						Trace.Fail("Shouldn't be here, there should be something wrong.");
						return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Unknown error.");
					}

					string postfix = requestUriTrimSlashes.Substring(webServiceUriTrimSlashes.Length).Trim('/');
					var postfixTokens = postfix.Split('/');

					if (postfixTokens == null || string.IsNullOrEmpty(postfixTokens[0]))
					{
						StaticLogger.LogWarn("No instance ID is specified.");
						return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified.");
					}

					if (!Guid.TryParse(postfixTokens[0], out instanceID))
					{
						StaticLogger.LogWarn("Bad instance ID: " + postfixTokens[0]);
						return DisplayConfigurationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is invalid.");
					}
				}

				// invoke callback function
				return this._displayConfigurationUpdateCallback(address, instanceID, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
