﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of keyboard mode configuration (update) (PUT /configuration/keyboard).
	/// </summary>
	internal sealed class KeyboardConfigurationUpdateAdapter : WebServiceAdapter<KeyboardConfigurationUpdateAdapter.KeyboardConfiguration, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/configuration/keyboard";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Enumerations

		/// <summary>
		/// Specifies the keyboard mode.
		/// </summary>
		public enum KeyboardMode
		{
			/// <summary>
			/// Signifies the keyboard mode is physical.
			/// </summary>
			[ParameterValue("physical")]
			Physical,

			/// <summary>
			/// Signifies the keyboard mode is virtual.
			/// </summary>
			[ParameterValue("virtual")]
			Virtual,
		}

		#endregion Enumerations

		#region Classes

		/// <summary>
		/// Represents the keyboard configuration.
		/// </summary>
		internal sealed class KeyboardConfiguration
		{
			/// <summary>
			/// Gets or sets the keyboard mode.
			/// </summary>
			[JsonProperty("Mode"), JsonConverter(typeof(AirFusion.Common.JsonConverters.EnumConverter))]
			internal KeyboardMode? Mode { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="KeyboardConfiguration"/> class.
			/// </summary>
			internal KeyboardConfiguration()
			{
				this.Mode = null;
			}

			/// <summary>
			/// Initializes a new instance of the <see cref="KeyboardConfiguration"/> class.
			/// </summary>
			/// <param name="mode">The keyboard mode.</param>
			internal KeyboardConfiguration(KeyboardMode mode)
			{
				this.Mode = mode;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Mode: " + (this.Mode.HasValue ? this.Mode.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="KeyboardConfigurationUpdateAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public KeyboardConfigurationUpdateAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Switches the keyboard mode on the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="mode">The keyboard mode.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> SetKeyboardModeAsync(IPAddress address, ushort port, KeyboardMode mode)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				new KeyboardConfiguration(mode));

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
