﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Reflection;

namespace AirFusion.Manager.FeatureHelper
{
	/// <summary>
	/// Represents the mix helper.
	/// </summary>
	public sealed partial class MixHelper
	{
		#region Statics Properties/Variables/Methods

		internal static IDictionary<UserDeviceManager, MixHelper> DeviceManagerMapping = new Dictionary<UserDeviceManager, MixHelper>();

		#endregion Statics Properties/Variables/Methods

		#region Properties

		/// <summary>
		/// Gets the mix service.
		/// </summary>
		internal MixService MixService { get; private set; }

		/// <summary>
		/// Gets or sets the version of application.
		/// </summary>
		public string ApplicationVersion
		{
			get
			{
				return this._applicationVersion;
			}
			set
			{
				this._applicationVersion = value;
				if (this._versionLocalRequestFilter != null)
				{
					this._versionLocalRequestFilter.ApplicationVersion = value;
				}
				if (this._versionRemoteRequestFilter != null)
				{
					this._versionRemoteRequestFilter.ApplicationVersion = value;
				}
				if (this._versionLocalResponseFilter != null)
				{
					this._versionLocalResponseFilter.ApplicationVersion = value;
				}
				if (this._versionRemoteResponseFilter != null)
				{
					this._versionRemoteResponseFilter.ApplicationVersion = value;
				}
			}
		}

		/// <summary>
		/// Gets or sets the minimum target version.
		/// </summary>
		public string MinimumTargetVersion
		{
			get
			{
				return this._minimumTargetVersion;
			}
			set
			{
				this._minimumTargetVersion = value;
				if (this._versionLocalRequestFilter != null)
				{
					this._versionLocalRequestFilter.MinimumTargetVersion = value;
				}
				if (this._versionRemoteRequestFilter != null)
				{
					this._versionRemoteRequestFilter.MinimumTargetVersion = value;
				}
				if (this._versionLocalResponseFilter != null)
				{
					this._versionLocalResponseFilter.MinimumTargetVersion = value;
				}
				if (this._versionRemoteResponseFilter != null)
				{
					this._versionRemoteResponseFilter.MinimumTargetVersion = value;
				}
			}
		}

		#endregion Properties

		#region Private/Protected/Internal Variables

		private UserDeviceManager _deviceManager;
		private string _applicationVersion;
		private string _minimumTargetVersion;
		private VersionLocalRequestFilter _versionLocalRequestFilter;
		private VersionRemoteRequestFilter _versionRemoteRequestFilter;
		private VersionLocalResponseFilter _versionLocalResponseFilter;
		private VersionRemoteResponseFilter _versionRemoteResponseFilter;

		//0609 Johnson add for record WebService Manager
		private WebServicesManager _webManager;

		#endregion Private/Protected/Internal Variables

		#region Events

		/// <summary>
		/// Occurs when a launch request is received.
		/// </summary>
		public event LaunchRequestedEventHandler LaunchRequested;

		#endregion Events

		#region Public Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="MixHelper"/> class.
		/// </summary>
		/// <param name="deviceManager">The device manager.</param>
		/// <param name="webManager">The web service manager.</param>
		internal MixHelper(UserDeviceManager deviceManager, WebServicesManager webManager)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (deviceManager == null)
				{
					throw new ArgumentNullException("deviceManager");
				}
				if (webManager == null)
				{
					throw new ArgumentNullException("webManager");
				}

				// debug assertion
				Debug.Assert(DeviceManagerMapping != null);

				// check prerequisite
				if (DeviceManagerMapping.ContainsKey(deviceManager))
				{
					throw new InvalidOperationException("The specified device manager already have a configuration helper.");
				}

				this._deviceManager = deviceManager;
				this._applicationVersion = null;
				this._minimumTargetVersion = null;
				this._versionLocalRequestFilter = null;
				this._versionRemoteRequestFilter = null;
				this._versionLocalResponseFilter = null;
				this._versionRemoteResponseFilter = null;

				DeviceManagerMapping.Add(deviceManager, this);

				//0609 Johnson add
				this._webManager = webManager;

				// configuration service
				this.MixService = new MixService(webManager);
				this.MixService.RegisterLaunchRequestCallback(this.OnLaunchRequested);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		//0609 Johnon add for Setting Support Version
		public void SettingSupportVersion(string applicationVersion, string minimumTargetVersion)
		{
			StaticLogger.TrackEnter();

			try
			{
				if (this.MixService != null)
				{
					// set versions
					this.ApplicationVersion = applicationVersion;
					this.MinimumTargetVersion = minimumTargetVersion;

					StaticLogger.LogTrace("ApplicationVersion : " + applicationVersion.ToString());
					StaticLogger.LogTrace("MinimumTargetVersion : " + minimumTargetVersion.ToString());

					// create filters
					this._versionLocalRequestFilter = new VersionLocalRequestFilter(this.ApplicationVersion, this.MinimumTargetVersion);
					this._versionRemoteRequestFilter = new VersionRemoteRequestFilter(this.ApplicationVersion, this.MinimumTargetVersion);
					this._versionLocalResponseFilter = new VersionLocalResponseFilter(this.ApplicationVersion, this.MinimumTargetVersion);
					this._versionRemoteResponseFilter = new VersionRemoteResponseFilter(this.ApplicationVersion, this.MinimumTargetVersion);

					if (this._webManager != null)
					{
						//Add Request filter in WebService Manager
						this._webManager.AddRequestFilter(1, this._versionRemoteRequestFilter);
						this._webManager.AddRequestFilter(2, this._versionLocalRequestFilter);

						//Add Response filter in WebService Manager
						this._webManager.AddResponseFilter(1, this._versionRemoteResponseFilter);
						this._webManager.AddResponseFilter(2, this._versionLocalResponseFilter);
					}
					else
					{
						StaticLogger.LogTrace("WebServicesManager is null, please double check.");
					}
				}
				else
				{
					StaticLogger.LogTrace("Not initialize MixHelp.");
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogTrace("SettingSupportVersion has some exception : " + ex.ToString());
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods

		#region Callbacks

		/// <summary>
		/// Called when a launch request is received.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="instanceID">The instance ID of device to be launched.</param>
		/// <returns>The response to the request.</returns>
		private LaunchRequestAdapter.WebServiceResponse OnLaunchRequested(IPAddress clientIP, Guid instanceID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("instanceID: " + instanceID.GetStringFormat());

				// debug assertion
				Debug.Assert(this._deviceManager != null);
				Debug.Assert(this._deviceManager.DeviceStore != null);

				// check prerequisite
				if (instanceID.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogError("The instance ID '" + instanceID.GetStringFormat() + "' is invalid.");
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + instanceID.GetStringFormat() + "' is invalid.");
				}
				if (instanceID.CompareTo(this._deviceManager.LocalDevice.InstanceID) == 0)
				{
					StaticLogger.LogError("The local device cannot be launched.");
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.Unauthorized, "The local device cannot be launched.");
				}

				// get device from device list
				var device = this._deviceManager.DeviceStore.GetDevice(instanceID);
				if (device == null)
				{
					StaticLogger.LogWarn("The device '" + instanceID.GetStringFormat() + "' is unrecognized.");
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotFound, "The device '" + instanceID.GetStringFormat() + "' is unrecognized. Please make sure you have both devices paired previously.");
				}

				// debug assertion
				Debug.Assert(device != null);

				// fire LaunchRequested event
				if (this.LaunchRequested != null)
				{
					StaticLogger.LogTrace("Fire LaunchRequested event...");
					CommonUtils.InvokeEvent<LaunchRequestedEventArgs>(new EventHandler<LaunchRequestedEventArgs>(this.LaunchRequested), this, new LaunchRequestedEventArgs(device));

					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
				else
				{
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}