﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Net;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the miscellaneous web services (/*).
	/// </summary>
	internal sealed class MixService
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle the launch request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="instanceID">The instance ID of the device to be launched.</param>
		public delegate LaunchRequestAdapter.WebServiceResponse LaunchCallback(IPAddress clientIP, Guid instanceID);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		private LaunchRequestAdapter _launchAdapter;
		private LaunchCallback _launchRequestCallback;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="MixService"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public MixService(WebServicesManager webServiceManager)
		{
			this._launchAdapter = new LaunchRequestAdapter(webServiceManager);
			this._launchAdapter.RegisterRequestCallback(this.OnLaunchAdapterRequestReceived);
		}

		/// <summary>
		/// Registers the callback function of launch request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterLaunchRequestCallback(LaunchCallback callback)
		{
			this._launchRequestCallback = callback;
		}

		#endregion Methods

		#region Callbacks

		/// <summary>
		/// Called when a launch request is received from a device.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private LaunchRequestAdapter.WebServiceResponse OnLaunchAdapterRequestReceived(IPAddress address, ushort port, LaunchRequestAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._launchRequestCallback == null)
				{
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered callback function.");
				}
				if (remoteRequest == null)
				{
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}

				// get instance ID
				if (remoteRequest.Request == null || remoteRequest.Request.InstanceID.HasValue == false)
				{
					return LaunchRequestAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "Please specify instance ID.");
				}

				// invoke callback function to get the response
				return this._launchRequestCallback(address, remoteRequest.Request.InstanceID.Value);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
