﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of wakeup (PUT /screencast/request).
	/// </summary>
	internal class ScreencastRequestAdapter : WebServiceAdapter<ScreencastRequestAdapter.ScreencastRequestInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/screencast/request";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the data of screencast request.
		/// </summary>
		internal class ScreencastRequestInfo
		{
			/// <summary>
			/// Gets or sets the instance ID of the requester.
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(AirFusion.Common.JsonConverters.GuidConverter))]
			public Guid? InstanceID { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="ScreencastRequestInfo"/> class.
			/// </summary>
			internal ScreencastRequestInfo()
			{
				this.InstanceID = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".InstanceID: " + (this.InstanceID.HasValue ? this.InstanceID.Value.GetStringFormat() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="ScreencastRequestAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public ScreencastRequestAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Sends a screencast request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="instanceID">The instance ID of the requester.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestAsync(IPAddress address, ushort port, Guid instanceID)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				new ScreencastRequestInfo() { InstanceID = instanceID });

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
