﻿using AirFusion.Manager.DeviceManager;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Manager.IncomingCall
{
	#region Enumerations for IIncomingCallManager

	///// <summary>
	///// Specifies the device to answer the phone
	///// </summary>
	//public enum AnswerPhoneDevice
	//{
	//	PC,
	//	Phone
	//}

	/// <summary>
	/// Specifies the decline type for the Decline() method of IIncomingCallManager
	/// </summary>
	public enum DeclineOption
	{
		/// <summary>
		/// Rejects the incoming call directly
		/// </summary>
		RejectIncomingCall,

		/// <summary>
		/// Rejects the incoming call and asks phone to send one short message to caller
		/// </summary>
		RejectCallWithMessage,

		/// <summary>
		/// Hangs up the phone call
		/// </summary>
		HangUpPhoneCall
	}

	#endregion Enumerations for IIncomingCallManager

	#region Event args for IIncomingCallManager's events

	/// <summary>
	/// Represents the event arguments of incoming call notification event
	/// </summary>
	public class ICCallReceivedEventArgs : EventArgs
	{
		#region Properties

		/// <summary>
		/// Gets particular device by the specified instance ID
		/// AFUserDevice includes:
		/// - DeviceInstanceId
		/// - DeviceName
		/// </summary>
		public AFUserDevice Device { get; private set; }

		/// <summary>
		/// Gets the phone number of caller
		/// </summary>
		public string PhoneNumber { get; private set; }

		/// <summary>
		/// Gets the name of caller
		/// </summary>
		public string CallerName { get; private set; }

		/// <summary>
		/// Gets the photo of caller
		/// </summary>
		public Bitmap CallerPhoto { get; private set; }

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="phone_number">The phone number of caller</param>
		/// <param name="caller_name">The name of caller</param>
		public ICCallReceivedEventArgs(AFUserDevice device, string phone_number, string caller_name)
		{
			this.Device = device;
			this.PhoneNumber = phone_number;
			this.CallerName = caller_name;
			this.CallerPhoto = null;
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="phone_number">The phone number of caller</param>
		/// <param name="caller_name">The name of caller</param>
		/// <param name="caller_photo">The photo of caller</param>
		public ICCallReceivedEventArgs(AFUserDevice device, string phone_number, string caller_name, Bitmap caller_photo)
		{
			this.Device = device;
			this.PhoneNumber = phone_number;
			this.CallerName = caller_name;
			this.CallerPhoto = caller_photo;
		}

		#endregion Properties
	}

	/// <summary>
	/// Represents the event arguments of incoming call state change event
	/// </summary>
	public class ICStateChangeEventArgs : EventArgs
	{
		/// <summary>
		/// Specifies the current state of one phone call
		/// </summary>
		public enum LineState
		{
			/// <summary>
			/// Phone received "Answer" request from PC and finished to answer the call
			/// </summary>
			AnswerOnPC,

			/// <summary>
			/// User answered the call on cell phone device
			/// </summary>
			AnswerOnPhone,

			/// <summary>
			/// Phone received "Decline" request from PC and finished to hang up the call/phone
			/// </summary>
			HangUpFromPC,

			/// <summary>
			/// User hanged up the call/phone on cell phone device
			/// </summary>
			HangUpFromPhone,

			/// <summary>
			/// The talk time on phone is updated
			/// </summary>
			UpdateTime,

			/// <summary>
			/// Unknown state
			/// </summary>
			Unknown
		}

		#region Properties

		/// <summary>
		/// Gets particular device by the specified instance ID
		/// AFUserDevice includes:
		/// - DeviceInstanceId
		/// </summary>
		public AFUserDevice Device { get; private set; }

		/// <summary>
		/// Gets the current state of phone call
		/// </summary>
		public LineState State { get; private set; }

		/// <summary>
		/// Gets the elapsed time after user answered the phone call
		/// </summary>
		public uint TalkTime { get; private set; }

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="call_state">The current state of incoming call</param>
		/// <param name="talk_time">The current talk time that phone shows</param>
		public ICStateChangeEventArgs(AFUserDevice device, LineState call_state, uint talk_time)
		{
			this.Device = device;
			this.State = call_state;
			this.TalkTime = talk_time;
		}

		#endregion Properties
	}

	#endregion Event args for IIncomingCallManager's events

	#region Delegates Event Handlers for IIncomingCallManager

	/// <summary>
	/// Represents the methods that will handle the event of receiving one incoming call notification
	/// </summary>
	/// <param name="sender"></param>
	/// <param name="e">ICCallReceivedEventArgs</param>
	public delegate void ICCallReceivedEventHandler(object sender, ICCallReceivedEventArgs e);

	/// <summary>
	/// Represents the methods that will handle the event of receiving the phone call state is changed
	/// - Answer, Hang up, Update talk time, ...
	/// </summary>
	/// <param name="sender"></param>
	/// <param name="e">ICStateChangeEventArgs</param>
	public delegate void ICStateChangedEventHandler(object sender, ICStateChangeEventArgs e);

	#endregion Delegates Event Handlers for IIncomingCallManager

	#region Class for the parameter of IIncomingCallManager's callback functions

	/// <summary>
	/// Represents the base class of the parameter of callback functions.
	/// </summary>
	public class ICBaseStatus
	{
		/// <summary>
		/// Specifies the types for the callback functions
		/// </summary>
		public enum StatusType
		{
			/// <summary>
			/// Signifies the operation is in progressing.
			/// </summary>
			InProgressMessage,

			/// <summary>
			/// Signifies the operation is done successfully.
			/// </summary>
			Completed_Success,

			/// <summary>
			/// Signifies the operation has ended with an error.
			/// </summary>
			Completed_Error

			// Add more error conditions if any.
		}

		/// <summary>
		/// Gets particular device by the specified instance ID
		/// AFUserDevice includes:
		/// - DeviceInstanceId
		/// </summary>
		public AFUserDevice Device { get; private set; }

		/// <summary>
		/// Gets the status type
		/// </summary>
		public StatusType Status { get; private set; }

		/// <summary>
		/// Gets the exception
		/// </summary>
		public Exception ICException { get; set; }

		/// <summary>
		/// Checks if the operation is done or not
		/// </summary>
		public bool IsCompleted
		{
			get
			{
				if ((this.Status == StatusType.Completed_Success) || (this.Status == StatusType.Completed_Error))
					return true;
				else
					return false;
			}
		}

		/// <summary>
		/// Checks if the operation is fault or not
		/// </summary>
		public bool IsFaulted
		{
			get
			{
				if (this.Status == StatusType.Completed_Error)
					return true;
				else
					return false;
			}
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="ICBaseStatus"/> class.
		/// </summary>
		/// <param name="dev"></param>
		/// <param name="s"></param>
		public ICBaseStatus(AFUserDevice dev, StatusType s)
		{
			this.Device = dev;
			this.Status = s;
		}
	}

	/// <summary>
	/// Represents the parameter class of AnswerCallback() function.
	/// </summary>
	public class AnswerStatus : ICBaseStatus
	{
		/// <summary>
		/// Specifies the error type for sending "Answer" request
		/// </summary>
		public enum Error
		{
			/// <summary>
			/// No error
			/// </summary>
			NO,

			/// <summary>
			/// PC's bluetooth is disabled
			/// </summary>
			PC_BT_Disabled,

			/// <summary>
			/// PC's paired bluetooth devices has no the phone device
			/// </summary>
			PC_BT_NotPaired,

			/// <summary>
			/// Phone's bluetooth is disabled
			/// </summary>
			Phone_BT_Disabled,

			/// <summary>
			/// Phone's paired bluetooth devices has no the PC device
			/// </summary>
			Phone_BT_NotPaired,

			/// <summary>
			/// There is no the mac address of phone
			/// </summary>
			No_Phone_BT_Address,

			/// <summary>
			/// Web service returns error
			/// </summary>
			WebService_Failed
		}

		/// <summary>
		/// Gets the error code
		/// </summary>
		public Error ErrorCode { get; private set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="AnswerStatus"/> class.
		/// </summary>
		/// <param name="device"></param>
		/// <param name="status"></param>
		internal AnswerStatus(AFUserDevice device, StatusType status)
			: base(device, status)
		{
			if (status == StatusType.Completed_Error)
				this.ErrorCode = Error.WebService_Failed;
			else
				this.ErrorCode = Error.NO;
		}

		internal AnswerStatus(AFUserDevice device, StatusType status, Error err)
			: base(device, status)
		{
			this.ErrorCode = err;
		}

		// [TBD] For IIncomingCallManager::Answer() method, it should have the errors below:
		// - Couldn't make the Bluetooth hands-free connection
	}

	/// <summary>
	/// Represents the parameter class of DeclineCallback() function.
	/// </summary>
	public class DeclineStatus : ICBaseStatus
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="DeclineStatus"/> class.
		/// </summary>
		/// <param name="device"></param>
		/// <param name="status"></param>
		internal DeclineStatus(AFUserDevice device, StatusType status)
			: base(device, status)
		{
		}

		/// <summary>
		/// Gets/Sets the decline option
		/// </summary>
		public DeclineOption Option { get; set; }

		// [TBD] For IIncomingCallManager::Decline() method, it should have the errors below
		// - Couldn't decline the incoming call (Maybe web service returns error)
		// - Couldn't hang up the phone (ex: Bluetooth API returns error)
	}

	#endregion Class for the parameter of IIncomingCallManager's callback functions

	#region Delegates Callback Functions for IIncomingCallManager

	/// <summary>
	/// Called when IIncomingCallManager::Answer() method would like to inform progress or error message
	/// </summary>
	/// <param name="status"></param>
	public delegate void AnswerCallback(AnswerStatus status);

	/// <summary>
	/// Called when IIncomingCallManager::Decline() method would like to inform progress or error message
	/// </summary>
	/// <param name="status"></param>
	public delegate void DeclineCallback(DeclineStatus status);

	#endregion Delegates Callback Functions for IIncomingCallManager

	#region IIncomingCallManager Interface

	/// <summary>
	/// Defines the methods/events of incoming call manager
	/// </summary>
	internal interface IIncomingCallManager
	{
		#region Events

		/// <summary>
		/// Occurs when PC receives a incoming call notification from phone
		/// </summary>
		event ICCallReceivedEventHandler CallReceived;

		/// <summary>
		/// Occurs when the state of phone call is changed
		/// </summary>
		event ICStateChangedEventHandler StateChanged;

		#endregion Events

		#region Methods

		/// <summary>
		/// Enables/Disables the feature of Answer phone call on PC
		/// </summary>
		/// <param name="enable">
		/// - True : Enable the feature of Answer phone call on PC
		/// - False: Disable the feature of Answer phone call on PC
		/// </param>
		void Enable(bool enable);

		/// <summary>
		/// Answers the incoming call on PC.
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="callback">Infrom progress/error message by running callback function</param>
		void Answer(AFUserDevice device, AnswerCallback callback);

		/// <summary>
		/// Rejects the incoming call or hangs up the phone call.
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="callback">Infrom progress/error message by running callback function</param>
		void Decline(AFUserDevice device, DeclineCallback callback);

		/// <summary>
		/// Hangs up the phone call
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="callback">Infrom progress/error message by running callback function</param>
		void HangUp(AFUserDevice device, DeclineCallback callback);

		#region [TBD] The Functions will be implemented in the future

		///// <summary>
		///// Rejects the incoming call and asks phone to send one short message to caller
		///// </summary>
		///// <param name="device">The phone device with incoming call</param>
		///// <param name="callback">Infrom progress/error message by running callback function</param>
		///// <param name="message">Short message</param>
		//void DeclineWithSms(AFUserDevice device, DeclineCallback callback, string message);

		///// <summary>
		///// Turns off the microphone of PC
		///// </summary>
		///// <param name="mute">
		///// - True : Turn off
		///// - False: Turn on
		///// </param>
		//void MuteMicrophone(bool mute);

		///// <summary>
		///// Switch the communication device
		///// </summary>
		///// <param name="device">
		///// - PC: Connect hands-free connection and it will return to call on PC
		///// - Phone: Disconnect hands-free connection and the voice will output from phone
		///// </param>
		///// <param name="target">If device == PC, the target is the smartphone</param>
		//void SwitchCommunDevice(AnswerPhoneDevice device, AFUserDevice target = null);

		#endregion [TBD] The Functions will be implemented in the future

		#endregion Methods
	}

	#endregion IIncomingCallManager Interface
}