﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.Manager.Notification
{
	/// <summary>
	/// TODO: please summary what the class is for.
	/// </summary>
	public sealed partial class NotificationManager
	{
		/// <summary>
		/// Represents the event arguments of a notification update.
		/// </summary>
		public sealed class AFNotificationItemUpdateEventArgs : EventArgs
		{
			/// <summary>
			/// Gets the notification ID
			/// </summary>
			public string NotificationKey { get; internal set; }

			/// <summary>
			/// Gets the device name
			/// </summary>
			public string DeviceName { get; internal set; }

			/// <summary>
			/// Gets the Toast item
			/// </summary>
			public ToastItem Toast { get; internal set; }
		}

		/// <summary>
		/// Represents the event arguments of a notification remove
		/// </summary>
		public sealed class AFNotificationItemRemoveEventArgs : EventArgs
		{
			/// <summary>
			/// Gets the notification ID
			/// </summary>
			public string NotificationKey { get; internal set; }
		}

		public sealed class AFNotificationIgnoreTimeOutedEventArgs : EventArgs
		{
			public string PackageName { get; internal set; }

			public Guid DeviceInstanceID { get; internal set; }

			/// <summary>
			/// Gets the notification ID
			/// </summary>
			public string NotificationKey { get; internal set; }

			public int RetryTime { get; internal set; }
		}

		#region Delegates

		/// <summary>
		/// Represents the method that will be invoked when a notification post request is received.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The event arguments.</param>
		public delegate void NotificationReceivedEventHandler(object sender, AFNotificationItemUpdateEventArgs args);

		/// <summary>
		/// Represents the method that will be invoked when a notification remove request is received.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The event arguments.</param>
		public delegate void NotificationRemovedEventHandler(object sender, AFNotificationItemRemoveEventArgs args);

		public delegate void NotificationIgnoreTimeOutedEventHandler(object sender, AFNotificationIgnoreTimeOutedEventArgs args);

		#endregion Delegates

		#region Event

		/// <summary>
		/// Occurs when a notification post request is received.
		/// </summary>
		public event NotificationReceivedEventHandler NotificationReceived;

		/// <summary>
		/// Occurs when a notification remove request is received.
		/// </summary>
		public event NotificationRemovedEventHandler NotificationRemoved;

		public event NotificationIgnoreTimeOutedEventHandler NotificationIgnoreTimeOuted;

		#endregion Event

		#region Protected/Private/Internal Variables

		private NotificationService _notificationService;
		private NotificationStore _notificationStore;
		private UserDeviceManager _userDeviceManager;

		#endregion Protected/Private/Internal Variables

		/// <summary>
		/// Initialize a new instance of the <see cref="NotificationManager"/> class.
		/// </summary>
		public NotificationManager()
		{
			this._notificationService = null;
			this._notificationStore = new NotificationStore();
		}

		/// <summary>
		/// Sets up the user device manager.
		/// </summary>
		/// <param name="userDeviceManager">The user device manager.</param>
		public void SetUserDeviceManager(UserDeviceManager userDeviceManager)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (userDeviceManager == null)
				{
					throw new ArgumentNullException("userDeviceManager");
				}

				this._userDeviceManager = userDeviceManager;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sets up the web service manager.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public void SetWebService(WebServicesManager webServiceManager)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (webServiceManager == null)
				{
					throw new ArgumentNullException("webServiceManager");
				}

				// notification service
				{
					this._notificationService = new NotificationService(webServiceManager);
					this._notificationService.RegisterNotificationPostRequestCallback(OnPostRequestReceived);
					this._notificationService.RegisterNotificationRemovalRequestCallback(OnRemoveRequestReceived);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#region Callbacks

		/// <summary>
		/// Called when a notification remove request is received.
		/// </summary>
		/// <param name="sourceIP">The remote device IP address</param>
		/// <param name="removeRequestUri">The remove request string</param>
		/// <returns>The response to the request</returns>
		private NotificationRemoveAdapter.WebServiceResponse OnRemoveRequestReceived(IPAddress sourceIP, string removeRequestUri)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (string.IsNullOrEmpty(removeRequestUri))
				{
					StaticLogger.LogError("The request data is unspecified.");
					return NotificationRemoveAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The request data is unspecified.");
				}

				if (string.IsNullOrEmpty(removeRequestUri))
				{
					StaticLogger.LogError("The instance ID is unspecified.");
					return NotificationRemoveAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID is unspecified.");
				}

				string notificationKey = NotificationKeyMaker.GetNotificationKeyByRemovalUri(removeRequestUri);

				//Check local database
				if (this.NotificationRemoved != null)
				{
					this._notificationStore.RemoveItemByKey(notificationKey);

					CommonUtils.BeginInvokeEvent<AFNotificationItemRemoveEventArgs>(
						new EventHandler<AFNotificationItemRemoveEventArgs>(this.NotificationRemoved),
						this,
						new AFNotificationItemRemoveEventArgs()
						{
							NotificationKey = notificationKey
						}
					);
				}

				return NotificationRemoveAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a notification post request is received.
		/// </summary>
		/// <param name="sourceIP">The remote device IP address</param>
		/// <param name="postData">The post request data</param>
		/// <returns>The response to the request</returns>
		private NotificationPostAdapter.WebServiceResponse OnPostRequestReceived(IPAddress sourceIP, NotificationPostAdapter.NotificationPostInfo postData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (postData == null)
				{
					StaticLogger.LogError("The request data is unspecified.");
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The request data is unspecified.");
				}
				if (!postData.DeviceInstanceID.HasValue)
				{
					StaticLogger.LogError("The instance ID is unspecified.");
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID is unspecified.");
				}
				if (postData.DeviceInstanceID.Value.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogError("The instance ID '" + postData.DeviceInstanceID.Value.GetStringFormat() + "' is invalid.");
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + postData.DeviceInstanceID.Value.GetStringFormat() + "' is invalid.");
				}

				if (this._notificationStore.GetDeviceIgnoreStatus(postData.DeviceInstanceID.ToString()) == DeviceIgnoreStatus.Ignore)
				{
					StaticLogger.LogInfo("All of the notification from this device should be ignored.");
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}

				//Check local database
				string notificationKey = String.Empty;
				NotificationHistoryItem item = this._notificationStore.GetOrCreateNewItem(postData, out notificationKey);
				List<string> queryIconIds = new List<string>();
				string notificatonImageIconPath = NotificationHelper.GetIconCacheFolder() + @"\" + item.IconId + ".png";
				if (!File.Exists(notificatonImageIconPath))
				{
					queryIconIds.Add(item.IconId);
				}

				if (postData.QuickActions != null && postData.QuickActions.Length > 0)
				{
					foreach (NotificationPostAdapter.NotificationQuickActionContent qa in postData.QuickActions)
					{
						if (!string.IsNullOrEmpty(qa.QAIcon) &&
							!File.Exists(NotificationHelper.GetIconCacheFolder() + @"\" + qa.QAIcon + ".png"))
						{
							queryIconIds.Add(qa.QAIcon);
						}
					}
				}

				if (queryIconIds.Count > 0)
				{
					StaticLogger.LogInfo("Query image required.");
					IPAddress targetDeviceIPAddress;
					ushort targetDevicePort;
					if (GetAddressAndPort(item.DeviceInstanceId.Value, out targetDeviceIPAddress, out targetDevicePort))
					{
						Task task = this.ImageRequestAsync(targetDeviceIPAddress, targetDevicePort, queryIconIds.ToArray());
						task.ContinueWith(delegate(Task prevTask)
						{
							if (prevTask.IsCompleted && prevTask.IsFaulted)
							{
								if (prevTask.Exception != null && prevTask.Exception.InnerException != null)
								{
									StaticLogger.LogError("Task Failed.\r\n" + CommonUtils.GetExceptionMessages(prevTask.Exception.InnerException));
								}
								else
								{
									StaticLogger.LogError("Task Failed.");
								}
							}
							else
							{
								StaticLogger.LogError("Task success.");
								if (this.NotificationReceived != null)
								{
									AFUserDevice userDevice = this._userDeviceManager.DeviceStore.GetDevice(item.DeviceInstanceId.Value);

									CommonUtils.BeginInvokeEvent<AFNotificationItemUpdateEventArgs>(
										new EventHandler<AFNotificationItemUpdateEventArgs>(this.NotificationReceived),
										this,
										new AFNotificationItemUpdateEventArgs()
										{
											NotificationKey = notificationKey,
											DeviceName = userDevice.DeviceName,
											Toast = ToastFactory.CreateToast(postData, notificatonImageIconPath)
										}
									);
								}
							}
						});
					}

					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
				else
				{
					if (this.NotificationReceived != null)
					{
						AFUserDevice userDevice = this._userDeviceManager.DeviceStore.GetDevice(item.DeviceInstanceId.Value);

						CommonUtils.BeginInvokeEvent<AFNotificationItemUpdateEventArgs>(
							new EventHandler<AFNotificationItemUpdateEventArgs>(this.NotificationReceived),
							this,
							new AFNotificationItemUpdateEventArgs()
							{
								NotificationKey = notificationKey,
								DeviceName = userDevice.DeviceName,
								Toast = ToastFactory.CreateToast(postData, notificatonImageIconPath)
							}
						);
					}

					StaticLogger.LogInfo("After invoke event...");

					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks

		#region Private Methods

		/// <summary>
		/// Requests icon from the target device
		/// </summary>
		/// <param name="ip">The ip address of the target device</param>
		/// <param name="port">The port of the target device</param>
		/// <param name="iconIds">The quired icon ids</param>
		/// <returns></returns>
		private async Task ImageRequestAsync(IPAddress ip, ushort port, params string[] iconIds)
		{
			try
			{
				StaticLogger.TrackEnter();

				foreach (string iconId in iconIds)
				{
					StaticLogger.LogInfo("Start query icon (" + iconId + ").");
					NotificationIconAdapter.WebServiceResponse response = await this._notificationService.ImageQueryAsync(ip, port, iconId).ConfigureAwait(continueOnCapturedContext: false);
					if (response.ResponseData != null && !string.IsNullOrEmpty(response.ResponseData.Icon))
					{
						StaticLogger.LogInfo("Query icon (" + iconId + ") successfully.");
						this.GenerateAndSaveImage(NotificationHelper.GetIconCacheFolder() + @"\" + iconId + ".png", response.ResponseData.Icon);
					}
					else
					{
						StaticLogger.LogInfo("Query qa icon (" + iconId + ") failed.");
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Check the address and the port of target device
		/// </summary>
		/// <param name="deviceID">The device id</param>
		/// <param name="address">The quried device IP address</param>
		/// <param name="port">The quried device port</param>
		/// <returns>A value to indicate if getting the IP address and the port successfully</returns>
		private bool GetAddressAndPort(Guid deviceID, out IPAddress address, out ushort port)
		{
			address = IPAddress.None;
			port = 0;

			if (deviceID == null)
			{
				return false;
			}

			if (this._userDeviceManager == null)
			{
				return false;
			}

			AFUserDevice userDevice = this._userDeviceManager.DeviceStore.GetDevice(deviceID);

			if (userDevice == null)
			{
				StaticLogger.LogError("Failed getting device.");
				return false;
			}
			if (userDevice.Address == IPAddress.None)
			{
				StaticLogger.LogWarn("The address of such device is unknown.");
				return false;
			}

			StaticLogger.LogInfo("Target Device: (" + userDevice.Address.ToString() + ", " + userDevice.WebServicePort + ")");
			address = userDevice.Address;
			port = userDevice.WebServicePort;

			return true;
		}

		/// <summary>
		/// TODO: please summary what the method is for.
		/// </summary>
		/// <param name="iconPath">TODO: please what's the parameter.</param>
		/// <param name="iconRawData">TODO: please what's the parameter.</param>
		private void GenerateAndSaveImage(string iconPath, string iconRawData)
		{
			Debug.Assert(iconRawData != null);
			Debug.Assert(iconPath != null);

			byte[] byteBuffer = Convert.FromBase64String(iconRawData);
			MemoryStream memoryStream = new MemoryStream(byteBuffer);
			memoryStream.Position = 0;

			Bitmap bitmapIcon = (Bitmap)Bitmap.FromStream(memoryStream);
			bitmapIcon.Save(iconPath);
			memoryStream.Close();
			memoryStream = null;
			byteBuffer = null;
		}

		#endregion Private Methods

		#region Public Methods

		/// <summary>
		/// Launch the notification on target device
		/// </summary>
		/// <param name="key">The notification id</param>
		public void NotificationLaunch(string key)
		{
			try
			{
				StaticLogger.TrackEnter();

				NotificationLaunchAdapter.NotificationLaunchInfo launchData = this._notificationStore.GetLaunchData(key);
				if (launchData != null)
				{
					IPAddress targetDeviceIPAddress;
					ushort targetDevicePort;
					if (GetAddressAndPort(this._notificationStore.GetDeviceInstanceID(key), out targetDeviceIPAddress, out targetDevicePort))
					{
						Task<NotificationLaunchAdapter.WebServiceResponse> task = this._notificationService.LaunchAsync(targetDeviceIPAddress, targetDevicePort, launchData);
						task.ContinueWith(delegate(Task t)
						{
							if (t.IsCompleted && t.IsFaulted)
							{
								if (t.Exception != null && t.Exception.InnerException != null)
								{
									StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));
								}
								else
								{
									StaticLogger.LogError("Error statusCode.");
								}
							}
						});
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Remove the notification on target device
		/// </summary>
		/// <param name="key">The notification id</param>
		public void NotificaitonRemove(string key)
		{
			try
			{
				StaticLogger.TrackEnter();

				string removeUri = this._notificationStore.GetRemoveUri(key);
				if (!string.IsNullOrEmpty(removeUri))
				{
					IPAddress targetDeviceIPAddress;
					ushort targetDevicePort;
					if (GetAddressAndPort(this._notificationStore.GetDeviceInstanceID(key), out targetDeviceIPAddress, out targetDevicePort))
					{
						Task<NotificationRemoveAdapter.WebServiceResponse> task = this._notificationService.RemoveAsync(targetDeviceIPAddress, targetDevicePort, removeUri);
						task.ContinueWith(delegate(Task t)
						{
							if (t.IsCompleted && t.IsFaulted)
							{
								if (t.Exception != null && t.Exception.InnerException != null)
								{
									StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));
								}
								else
								{
									StaticLogger.LogError("Error statusCode.");
								}
							}
						});
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Forbid (Ignore) the notification to show on the PC
		/// </summary>
		/// <param name="key">The notification id</param>
		public void NotificationIgnore(string _packageName, Guid _deviceInstanceID, string _key, int _retryTime)
		{
			try
			{
				string packageName = _packageName;
				Guid deviceInstanceID = _deviceInstanceID;

				StaticLogger.TrackEnter();

				if (this._userDeviceManager == null)
				{
					StaticLogger.LogError("UserDeviceManager is null.");
					return;
				}

				if (_retryTime == 0)
				{
					NotificationHistoryItem item = this._notificationStore.GetCacheItem(_key);

					if (item == null)
					{
						StaticLogger.LogError("Could not get local cache notification item.");
						return;
					}

					packageName = item.PackageName;
					deviceInstanceID = this._notificationStore.GetDeviceInstanceID(_key);
				}

				if (this._userDeviceManager.LocalDevice == null ||
					this._userDeviceManager.LocalDevice.InstanceID == Guid.Empty)
				{
					StaticLogger.LogError("Could not get local device instanceID.");
					return;
				}

				string ignoreRequestUri = "/" + this._userDeviceManager.LocalDevice.InstanceID.ToString().ToUpperInvariant()
										+ "/" + packageName;

				if (!string.IsNullOrEmpty(ignoreRequestUri))
				{
					IPAddress targetDeviceIPAddress;
					ushort targetDevicePort;
					if (GetAddressAndPort(deviceInstanceID, out targetDeviceIPAddress, out targetDevicePort))
					{
						Task<NotificationIgnoreAdapter.WebServiceResponse> task = this._notificationService.ConfigureIgnoreAsync(targetDeviceIPAddress, targetDevicePort, ignoreRequestUri);
						task.ContinueWith(delegate(Task t)
						{
							if (t.IsCompleted && t.IsFaulted)
							{
								if (t.Exception != null && t.Exception.InnerException != null)
								{
									StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));

									AggregateException ae = t.Exception as AggregateException;

									foreach (var e in ae.Flatten().InnerExceptions)
									{
										if (e is TimeoutException)
										{
											StaticLogger.LogError("NotificationIgnore timeout'" + this + "': " + e.Message);

											try
											{
												if (this.NotificationIgnoreTimeOuted != null)
												{
													this.NotificationIgnoreTimeOuted(this, new AFNotificationIgnoreTimeOutedEventArgs()
													{
														PackageName = packageName,
														DeviceInstanceID = deviceInstanceID,
														NotificationKey = _key,
														RetryTime = _retryTime,
													});
												}
											}
											catch (Exception ex)
											{
												string tmp = ex.ToString();
											}

											return;
										}
									}
								}
								else
								{
									StaticLogger.LogError("Error statusCode.");
								}
							}
						});
					}
				}
			}
			catch (TimeoutException ex)
			{
				StaticLogger.LogError("NotificationIgnore timeout");
				throw ex;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Launch the specific quick action of the notification on target device
		/// </summary>
		/// <param name="key">The notification id</param>
		/// <param name="qaId">The quick action id</param>
		public void QuickActionLaunch(string key, int qaId)
		{
			try
			{
				StaticLogger.TrackEnter();

				NotificationQuickActionAdapter.NotificationQuickActionInfo requestData = this._notificationStore.GetQuickActionData(key, qaId);
				if (requestData != null)
				{
					IPAddress targetDeviceIPAddress;
					ushort targetDevicePort;
					if (GetAddressAndPort(this._notificationStore.GetDeviceInstanceID(key), out targetDeviceIPAddress, out targetDevicePort))
					{
						Task<NotificationQuickActionAdapter.WebServiceResponse> task = this._notificationService.QuickActionLaunchAsync(targetDeviceIPAddress, targetDevicePort, requestData);
						task.ContinueWith(delegate(Task t)
						{
							if (t.IsCompleted && t.IsFaulted)
							{
								if (t.Exception != null && t.Exception.InnerException != null)
								{
									StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));
								}
								else
								{
									StaticLogger.LogError("Error statusCode.");
								}
							}
						});
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Remove the notification item on local datebase
		/// </summary>
		/// <param name="key">The notification id</param>
		public void RemoveItemOnLocalDatabase(string key)
		{
			try
			{
				StaticLogger.TrackEnter();

				this._notificationStore.RemoveItemByKey(key);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Ignore / Permit the notification from the device.
		/// </summary>
		/// <param name="deviceId">The device instance ID.</param>
		/// <param name="status">The device ignore status.</param>
		public void ConfigureDeviceNotificationStatus(string deviceId, DeviceIgnoreStatus status)
		{
			try
			{
				StaticLogger.TrackEnter();

				this._notificationStore.SetDeviceIgnoreStatus(deviceId, status);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods
	}

	#region Disposable

	public sealed partial class NotificationManager : IDisposable
	{
		private bool _disposed = false;

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources.
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		private void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					//do something
				}

				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		// Implement IDisposable.
		// Do not make this method virtual.
		// A derived class should not be able to override this method.
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}
	}

	#endregion Disposable
}