﻿using AirFusion.Common;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;

namespace AirFusion.Manager.Notification
{
	/// <summary>
	/// Provides the functionality of NotificationStore.
	/// </summary>
	public sealed class NotificationStore
	{
		private ConcurrentDictionary<string, NotificationHistoryItem> _connPool;
		private ConcurrentDictionary<string, DeviceIgnoreStatus> _deviceIgnorePool;

		/// <summary>
		/// Initializes the <see cref="NotificationStore" /> instance.
		/// </summary>
		public NotificationStore()
		{
			this._connPool = new ConcurrentDictionary<string, NotificationHistoryItem>();
			this._deviceIgnorePool = new ConcurrentDictionary<string, DeviceIgnoreStatus>();
		}

		internal DeviceIgnoreStatus GetDeviceIgnoreStatus(string deviceId)
		{
			try
			{
				StaticLogger.TrackEnter();

				DeviceIgnoreStatus status = DeviceIgnoreStatus.Permit;

				if (this._deviceIgnorePool.TryGetValue(deviceId, out status))
				{
				}
				else // get status from pool fail
				{
					status = DeviceIgnoreStatus.Permit;
				}

				return status;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		internal void SetDeviceIgnoreStatus(string deviceId, DeviceIgnoreStatus status)
		{
			try
			{
				StaticLogger.TrackEnter();

				DeviceIgnoreStatus resultStatus = this._deviceIgnorePool.AddOrUpdate(deviceId, status,
					(key, existingStatus) =>
					{
						existingStatus = status;
						return existingStatus;
					});
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets or creates a new notification history record
		/// </summary>
		/// <param name="data">The posted notification packet</param>
		/// <param name="notificationKey">The notification key generated by the posted notification packet</param>
		internal NotificationHistoryItem GetOrCreateNewItem(NotificationPostAdapter.NotificationPostInfo data, out string notificationKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(data != null);
				Debug.Assert(this._connPool != null);

				notificationKey = NotificationKeyMaker.GetNotificationKeyByPostData(data);
				NotificationHistoryItem newItem = new NotificationHistoryItem(data);
				this._connPool.AddOrUpdate(notificationKey, newItem, (key, existingVal) => newItem);

				return newItem;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Removes a notification from pool by notification key
		/// </summary>
		/// <param name="notificationKey">The notification key</param>
		internal void RemoveItemByKey(string notificationKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(notificationKey != null);
				Debug.Assert(this._connPool != null);

				NotificationHistoryItem removedItem;
				if (!this._connPool.TryRemove(notificationKey, out removedItem))
				{
					StaticLogger.LogInfo("Remove item failed.");
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the remove request uri by notification key
		/// </summary>
		/// <param name="notificationKey">The notification key</param>
		/// <returns>An uri used to remove a notification</returns>
		internal string GetRemoveUri(string notificationKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(notificationKey != null);
				Debug.Assert(this._connPool != null);

				NotificationHistoryItem item;
				if (this._connPool.TryGetValue(notificationKey, out item))
				{
					//UserId, PackageName, Id must have values
					string uri =
						   "/" + item.UserId.ToString() +
						   "/" + item.PackageName +
						   "/" + item.Id.ToString();

					//Tag may be null (empty)
					if (item.Tag != null)
					{
						uri += "/" + item.Tag;
					}

					return uri;
				}
				else
				{
					StaticLogger.LogError("Get cache item failed.");
					return null;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets a notification launch request data structure by notification key
		/// </summary>
		/// <param name="notificationKey">The notification key</param>
		/// <returns>A structure used to launch a specific notification</returns>
		internal NotificationLaunchAdapter.NotificationLaunchInfo GetLaunchData(string notificationKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(notificationKey != null);
				Debug.Assert(this._connPool != null);

				NotificationHistoryItem item;
				if (this._connPool.TryGetValue(notificationKey, out item))
				{
					NotificationLaunchAdapter.NotificationLaunchInfo launchData = new NotificationLaunchAdapter.NotificationLaunchInfo();
					launchData.Id = item.Id;
					launchData.UserId = item.UserId;
					launchData.Tag = item.Tag;
					launchData.PackageName = item.PackageName;
					launchData.DeviceInstanceID = item.DeviceInstanceId;

					return launchData;
				}
				else
				{
					StaticLogger.LogError("Get cache item failed.");
					return null;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets a execution request structure of notification quick action by notification key
		/// </summary>
		/// <param name="notificationKey">The notification key</param>
		/// <param name="qaId">The quick action id</param>
		/// <returns>A structure used to execute a specific notification quick action</returns>
		internal NotificationQuickActionAdapter.NotificationQuickActionInfo GetQuickActionData(string notificationKey, int qaId)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(notificationKey != null);
				Debug.Assert(this._connPool != null);

				NotificationHistoryItem item;
				if (this._connPool.TryGetValue(notificationKey, out item))
				{
					NotificationQuickActionAdapter.NotificationQuickActionInfo qaData = new NotificationQuickActionAdapter.NotificationQuickActionInfo();
					qaData.Id = item.Id;
					qaData.UserId = item.UserId;
					qaData.Tag = item.Tag;
					qaData.PackageName = item.PackageName;
					qaData.QuickActionId = qaId;

					return qaData;
				}
				else
				{
					StaticLogger.LogError("Get cache item failed.");
					return null;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the notificaiton cache item
		/// </summary>
		/// <param name="notificationKey">The notification key</param>
		internal NotificationHistoryItem GetCacheItem(string notificationKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(notificationKey != null);
				Debug.Assert(this._connPool != null);

				NotificationHistoryItem item;
				if (this._connPool.TryGetValue(notificationKey, out item))
				{
					return item;
				}
				else
				{
					StaticLogger.LogError("Get cache item failed.");
					return null;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the device ID by notification key
		/// </summary>
		/// <param name="notificationKey">The notification key</param>
		/// <returns>A Guid format</returns>
		internal Guid GetDeviceInstanceID(string notificationKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(notificationKey != null);
				Debug.Assert(this._connPool != null);

				NotificationHistoryItem item;
				if (this._connPool.TryGetValue(notificationKey, out item))
				{
					return item.DeviceInstanceId.Value;
				}
				else
				{
					StaticLogger.LogError("Get cache item failed.");
					return Guid.Empty;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Removes all the items in the notification pool
		/// </summary>
		internal void ClearDatabase()
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._connPool != null);

				this._connPool.Clear();
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
	}

	/// <summary>
	/// Represents a notification history item
	/// </summary>
	public class NotificationHistoryItem
	{
		/// <summary>
		/// Gets the notification Id value.
		/// </summary>
		public int? Id { get; internal set; }

		/// <summary>
		/// Gets the notification UserId value.
		/// </summary>
		public int? UserId { get; internal set; }

		/// <summary>
		/// Gets the notification Tag value.
		/// </summary>
		public string Tag { get; internal set; }

		/// <summary>
		/// Gets the apk name which generates the notification.
		/// </summary>
		public string PackageName { get; internal set; }

		/// <summary>
		/// Gets the app apk version on android
		/// </summary>
		public int? PackageVersion { get; internal set; }

		/// <summary>
		/// Gets the icon id
		/// </summary>
		public string IconId { get; internal set; }

		/// <summary>
		/// Gets the device IP address
		/// </summary>
		public IPAddress ClientIPAddress { get; internal set; }

		/// <summary>
		/// Gets the instance ID of the source device that posted the notification.
		/// </summary>
		public Guid? DeviceInstanceId { get; internal set; }

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		internal NotificationHistoryItem()
		{
			this.Id = null;
			this.UserId = null;
			this.Tag = null;
			this.PackageName = null;
			this.PackageVersion = null;
			this.IconId = null;
			this.DeviceInstanceId = null;
		}

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		internal NotificationHistoryItem(NotificationPostAdapter.NotificationPostInfo postData)
		{
			this.Id = postData.Id;
			this.UserId = postData.UserId;
			this.Tag = postData.Tag;
			this.PackageName = postData.PackageName;
			this.DeviceInstanceId = postData.DeviceInstanceID;
			this.PackageVersion = postData.PackageVersion;

			if (!string.IsNullOrEmpty(postData.AppIconId))
			{
				this.IconId = postData.AppIconId;
			}
			else if (!string.IsNullOrEmpty(postData.SmallIconId))
			{
				this.IconId = postData.SmallIconId;
			}
			else
			{
				this.IconId = postData.LargeIconId;
			}
		}
	}
}
