﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Provides the web service of notification icon request (PC => Phone)
	/// (GET /notification/icon/[PackageName]/[PackageVersion]/[Icon resid]/[json | png]).
	/// </summary>
	internal sealed class NotificationIconAdapter : WebServiceAdapter<DBNull, NotificationIconAdapter.NotificationQueryIconResponse>
	{
		#region Definitions

		internal const string WebServiceUri = "/notification/icon";
		internal static HttpMethod WebServiceMethod = HttpMethod.Get;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Response data of the removal request.
		/// </summary>
		public sealed class NotificationQueryIconResponse
		{
			/// <summary>
			/// Gets the icon string
			/// </summary>
			[JsonProperty("icon")]
			public string Icon { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="NotificationQueryIconResponse"/> class.
			/// </summary>
			public NotificationQueryIconResponse()
			{
				this.Icon = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Icon: " + this.Icon);

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationIconAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public NotificationIconAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Sends a notification icon request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="iconRequestUri">
		/// The required icon string to specify which notification to launch.
		/// The request string should follow the format.
		/// /notification/icon/[PackageName]/[PackageVersion]/[Icon_resid]/[json | png].
		/// </param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestImageAsync(IPAddress address, ushort port, string iconRequestUri)
		{
			// Seoker comment: the request URI should be composed in this object, the responsibility doesn't belong to upper-layer.

			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + iconRequestUri).Uri);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
