﻿using AirFusion.Common;
using AirFusion.Common.JsonConverters;
using Newtonsoft.Json;
using System;
using System.Net.Http;
using System.Text;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of notification post from Android side (Phone => PC)
	/// (POST /notification/).
	/// </summary>
	internal sealed class NotificationPostAdapter : WebServiceAdapter<NotificationPostAdapter.NotificationPostInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/notification";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the (large) content view data of notification post.
		/// </summary>
		public sealed class NotificationContentView
		{
			#region Properties

			/// <summary>
			/// Gets the style of the content view.
			/// Accept value : 0, 1, 2, 3, 4, 5
			/// </summary>
			[JsonProperty("style")]
			public int? Style { get; set; }

			/// <summary>
			/// Gets the notification title.
			/// Accept value : string or null.
			/// 1 line in style 0; max 3 lines in other styles
			/// </summary>
			[JsonProperty("title")]
			public string Title { get; set; }

			/// <summary>
			/// Gets the middle notification content.
			/// Accept value : string or null.
			/// 1 line in style 0, 1, 2; max 8 lines in style 3
			/// </summary>
			[JsonProperty("text2")]
			public string MiddleText { get; set; }

			/// <summary>
			/// Gets the bottom notification content.
			/// Accept value : string or null.
			/// </summary>
			[JsonProperty("text")]
			public string BottomText { get; set; }

			/// <summary>
			/// Gets the notification info (texts on right-bottom side)
			/// Accept value : string or null.
			/// </summary>
			[JsonProperty("info")]
			public string Info { get; set; }

			/// <summary>
			/// Gets the notification inbox content.
			/// Only support on style 4
			/// </summary>
			[JsonProperty("inbox_texts")]
			public string[] InboxTexts { get; set; }

			/// <summary>
			/// Gets the notification custom content.
			/// Only support on style 5
			/// </summary>
			[JsonProperty("c_texts")]
			public string[] CustomTexts { get; set; }

			#endregion Properties

			/// <summary>
			/// Initializes a new instance of the <see cref="NotificationContentView"/> class.
			/// </summary>
			internal NotificationContentView()
			{
				this.Style = null;
				this.Title = null;
				this.MiddleText = null;
				this.BottomText = null;
				this.Info = null;
				this.InboxTexts = null;
				this.CustomTexts = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Style: " + (this.Style.HasValue ? this.Style.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Title: " + (!string.IsNullOrEmpty(this.Title) ? this.Title : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".MiddleText: " + (!string.IsNullOrEmpty(this.MiddleText) ? this.MiddleText : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".BottomText: " + (!string.IsNullOrEmpty(this.BottomText) ? this.BottomText : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Info: " + (!string.IsNullOrEmpty(this.Info) ? this.Info : "N/A"));

				if (this.InboxTexts != null)
				{
					builder.AppendLine(this.GetType().GetFullName() + ".InboxTexts: ");

					foreach (string content in this.InboxTexts)
					{
						builder.AppendLine(content);
					}
				}

				if (this.CustomTexts != null)
				{
					builder.AppendLine(this.GetType().GetFullName() + ".CustomTexts: ");

					foreach (string content in this.CustomTexts)
					{
						builder.AppendLine(content);
					}
				}

				return builder.ToString().TrimEnd();
			}
		}

		/// <summary>
		/// Represents the quick action content of notification post.
		/// </summary>
		public sealed class NotificationQuickActionContent
		{
			/// <summary>
			/// Gets the quick action title
			/// </summary>
			[JsonProperty("qa_title")]
			public string QATitle { get; set; }

			/// <summary>
			/// Gets if the action intent will pop-up another display
			/// </summary>
			[JsonProperty("qa_is_activity")]
			public bool? QAActivityPopup { get; set; }

			/// <summary>
			/// Gets the quick action icon
			/// </summary>
			[JsonProperty("qa_icon")]
			public string QAIcon { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="NotificationQuickActionContent"/> class.
			/// </summary>
			internal NotificationQuickActionContent()
			{
				this.QATitle = null;
				this.QAIcon = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".QATitle: " + (!string.IsNullOrEmpty(this.QATitle) ? this.QATitle : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".QAIcon: " + (!string.IsNullOrEmpty(this.QAIcon) ? this.QAIcon : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".QAActivityPopup: " + ((this.QAActivityPopup.HasValue) ? this.QAActivityPopup.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		/// <summary>
		/// Represents the request data of notification post.
		/// </summary>
		public sealed class NotificationPostInfo
		{
			#region Properties

			/// <summary>
			/// Gets the instance ID of the device.
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(GuidConverter))]
			public Guid? DeviceInstanceID { get; set; }

			/// <summary>
			/// Gets the notification id.
			/// </summary>
			[JsonProperty("id")]
			public int? Id { get; set; }

			/// <summary>
			/// Gets the user id.
			/// </summary>
			[JsonProperty("user_id")]
			public int? UserId { get; set; }

			/// <summary>
			/// Gets the notification tag.
			/// </summary>
			[JsonProperty("tag")]
			public string Tag { get; set; }

			/// <summary>
			/// Gets the package name of apk which generates the notification.
			/// </summary>
			[JsonProperty("package_name")]
			public string PackageName { get; set; }

			/// <summary>
			/// Gets the apk package version.
			/// </summary>
			[JsonProperty("package_version")]
			public int? PackageVersion { get; set; }

			/// <summary>
			/// Gets the app name of apk which generates the notification.
			/// </summary>
			[JsonProperty("app_name")]
			public string AppName { get; set; }

			/// <summary>
			/// Gets the ID of small icon.
			/// </summary>
			[JsonProperty("small_icon_md5")]
			public string SmallIconId { get; internal set; }

			/// <summary>
			/// Gets the ID of application icon.
			/// </summary>
			[JsonProperty("app_icon_md5")]
			public string AppIconId { get; internal set; }

			/// <summary>
			/// Gets the ID of large icon.
			/// </summary>
			[JsonProperty("large_icon_md5")]
			public string LargeIconId { get; internal set; }

			/// <summary>
			/// Gets a value indicating if the notification intent will bring to an activity.
			/// </summary>
			[JsonProperty("is_activity")]
			public bool? ActivityPopup { get; internal set; }

			/// <summary>
			/// Gets the property if the notification could be removed.
			/// </summary>
			[JsonProperty("removable")]
			public bool? Removable { get; internal set; }

			/// <summary>
			/// Gets the array of quick action content.
			/// </summary>
			[JsonProperty("quick_actions")]
			public NotificationQuickActionContent[] QuickActions { get; internal set; }

			/// <summary>
			/// Gets the big content view.
			/// </summary>
			[JsonProperty("big_content_view")]
			public NotificationContentView BigContentView { get; internal set; }

			/// <summary>
			/// Gets the content view.
			/// </summary>
			[JsonProperty("content_view")]
			public NotificationContentView ContentView { get; internal set; }

			#endregion Properties

			/// <summary>
			/// Initializes a new instance of the <see cref="NotificationPostInfo"/> class.
			/// </summary>
			internal NotificationPostInfo()
			{
				this.DeviceInstanceID = null;
				this.Id = null;
				this.UserId = null;
				this.Tag = null;
				this.PackageName = null;
				this.PackageVersion = null;
				this.AppName = null;
				this.SmallIconId = null;
				this.AppIconId = null;
				this.LargeIconId = null;
				this.ActivityPopup = null;
				this.Removable = null;
				this.QuickActions = null;
				this.BigContentView = new NotificationContentView();
				this.ContentView = new NotificationContentView();
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".DeviceInstanceID: " + (this.DeviceInstanceID.HasValue ? this.DeviceInstanceID.Value.GetStringFormat() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".ID: " + (this.Id.HasValue ? this.Id.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".UserID: " + (this.UserId.HasValue ? this.UserId.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Tag: " + (!string.IsNullOrEmpty(this.Tag) ? this.Tag : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".PackageName: " + (!string.IsNullOrEmpty(this.PackageName) ? this.PackageName : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".PackageVersion: " + (this.PackageVersion.HasValue ? this.PackageVersion.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".AppName: " + (!string.IsNullOrEmpty(this.AppName) ? this.AppName : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".SmallIconId: " + (!string.IsNullOrEmpty(this.SmallIconId) ? this.SmallIconId : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".AppIconId: " + (!string.IsNullOrEmpty(this.AppIconId) ? this.AppIconId : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".LargeIconId: " + (!string.IsNullOrEmpty(this.LargeIconId) ? this.LargeIconId : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".ActivityPopup: " + (this.ActivityPopup.HasValue ? this.ActivityPopup.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Removable: " + (this.Removable.HasValue ? this.Removable.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".BigContentView: " + (this.BigContentView != null ? "\r\n" + this.BigContentView.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".ContentView: " + (this.ContentView != null ? "\r\n" + this.ContentView.ToString() : "N/A"));

				if (this.QuickActions != null)
				{
					builder.AppendLine(this.GetType().GetFullName() + ".QuickActions: ");
					foreach (NotificationQuickActionContent action in this.QuickActions)
					{
						builder.AppendLine(action.ToString());
					}
				}

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationPostAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public NotificationPostAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		#endregion Methods
	}
}
