﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the template of a web service.
	/// </summary>
	/// <typeparam name="TRequest">The type of request object.</typeparam>
	/// <typeparam name="TResponse">The type of response object.</typeparam>
	public partial class WebServiceAdapter<TRequest, TResponse>
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle the remote request.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response.</returns>
		public delegate WebServiceResponse RequestReceivedCallback(IPAddress address, ushort port, WebServiceRequest remoteRequest);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		private WebServicesManager _webServiceManager;
		private RequestReceivedCallback _requestCallback;

		#endregion Private/Protected/Internal Variables

		#region Properteis

		/// <summary>
		/// Gets or sets the scheme, e.g. http
		/// </summary>
		protected string Scheme { get; set; }

		/// <summary>
		/// Gets or sets the timeout.
		/// </summary>
		protected TimeSpan? Timeout { get; set; }

		#endregion Properteis

		#region Methods

		/// <summary>
		/// Prevents a default instance of the <see cref="WebServiceAdapter{TRequest, TResponse}"/> class from being created.
		/// </summary>
		private WebServiceAdapter()
		{
			this._webServiceManager = null;
			this._requestCallback = null;
			this.Scheme = "http";
			this.Timeout = TimeSpan.FromMilliseconds(8000);
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="WebServiceAdapter{TRequest, TResponse}"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public WebServiceAdapter(WebServicesManager webServiceManager)
			: this()
		{
			this._webServiceManager = webServiceManager;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="WebServiceAdapter{TRequest, TResponse}"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		/// <param name="timout">The custom web service timeout.</param>
		public WebServiceAdapter(WebServicesManager webServiceManager, TimeSpan? timout)
			: this(webServiceManager)
		{
			this.Timeout = timout;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="WebServiceAdapter{TRequest, TResponse}"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		/// <param name="prefix">The prefix URI to listen.</param>
		/// <param name="method">The method.</param>
		public WebServiceAdapter(WebServicesManager webServiceManager, string prefix, HttpMethod method = null)
			: this(webServiceManager)
		{
			this._webServiceManager.RegisterPrefix(prefix, this.GetResponse, WebServicesManager.ExpectedPayloadContentType.StringContent, method);
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="WebServiceAdapter{TRequest, TResponse}"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		/// <param name="prefix">The prefix.</param>
		/// <param name="timeout">The custom web service timeout.</param>
		/// <param name="method">The method.</param>
		public WebServiceAdapter(WebServicesManager webServiceManager, string prefix, TimeSpan? timeout, HttpMethod method = null)
			: this(webServiceManager, timeout)
		{
			this._webServiceManager.RegisterPrefix(prefix, this.GetResponse, WebServicesManager.ExpectedPayloadContentType.StringContent, method);
		}

		/// <summary>
		/// Registers the callback of request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public virtual void RegisterRequestCallback(RequestReceivedCallback callback)
		{
			this._requestCallback = callback;
		}

		/// <summary>
		/// Gets the response.
		/// </summary>
		/// <param name="request">The request.</param>
		/// <returns>The string-format response.</returns>
		protected virtual HttpResponseMessage GetResponse(WebServicesManager.HttpRemoteRequest request)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("request: " + (request != null ? "\r\n" + request.ToString() : "N/A"));

				Trace.Assert(request.RequestMessage != null, "request.RequestMessage = null");
				Trace.Assert(request.RequestMessage.Method != null, "request.RequestMessage.Method = null");
				Trace.Assert(request.RequestMessage.RequestUri != null, "request.RequestMessage.RequestUri = null");

				TRequest requestPayload = default(TRequest);

				if (request.RequestMessage.PayloadContent != null && request.RequestMessage.PayloadContent is string)
				{
					var requestBody = request.RequestMessage.PayloadContent as string;

					// find first left brace & last right brace
					int leftBrace = requestBody.IndexOf('{');
					int rightBrace = requestBody.LastIndexOf('}');

					if (leftBrace >= 0 && rightBrace >= leftBrace)
					{
						requestPayload = CommonUtils.DeserializeFromJson<TRequest>(requestBody.Substring(leftBrace, rightBrace - leftBrace + 1));
					}
				}

				// TODO:
				// read the content into a buffer in ReadContentAsync()

				var responseMessage = new HttpResponseMessage();
				responseMessage.Headers.ConnectionClose = true;

				// response
				if (this._requestCallback != null)
				{
					try
					{
						var webServiceRequest = WebServiceRequest.CreateRemoteRequest(request.RequestMessage, requestPayload);
						var response = this._requestCallback(request.ClientIP, request.Port, webServiceRequest);
						if (response == null)
						{
							Trace.Fail("The response of '" + request.RequestMessage.RequestUri + "' is null.");

							// 500 InternalServerError
							responseMessage.StatusCode = HttpStatusCode.InternalServerError;
							responseMessage.Content = new StringContent("No response from the callback function.", Encoding.UTF8, "text/plain");
						}
						else
						{
							responseMessage.StatusCode = response.StatusCode;

							if (response.IsMessageResponse)
							{
								responseMessage.Content = new StringContent(response.Message, Encoding.UTF8, "text/plain");
							}
							else
							{
								responseMessage.Content = new StringContent(CommonUtils.SerializeAsJson(response.ResponseData), Encoding.UTF8, "application/json");
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("Unhandled exception of web service '" + request.RequestMessage.Method + " " + request.RequestMessage.RequestUri + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));

						// 500 InternalServerError
						responseMessage.StatusCode = HttpStatusCode.InternalServerError;
						responseMessage.Content = new StringContent("Unhandled exception, message: " + ex.Message + "\r\nStack trace:\r\n" + ex.StackTrace, Encoding.UTF8, "text/plain");
					}
				}
				else
				{
					// 501 NotImplemented
					responseMessage.StatusCode = HttpStatusCode.NotImplemented;
					responseMessage.Content = new StringContent("No registered callback function.", Encoding.UTF8, "text/plain");
				}

				return responseMessage;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends HTTP request asynchronously.
		/// </summary>
		/// <param name="request">The request.</param>
		/// <param name="timeout">The timeout.</param>
		/// <returns></returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public virtual async Task<WebServiceResponse> SendRequestAsync(WebServiceRequest request, TimeSpan? timeout = null)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (request == null)
				{
					throw new ArgumentNullException("request");
				}

				var responseMessage = await this._webServiceManager.SendRequestAsync(request.RequestMessage, WebServicesManager.ExpectedPayloadContentType.StringContent, timeout);
				string stringResponse = string.Empty;
				if (responseMessage.Content is StringContent)
				{
					stringResponse = await (responseMessage.Content as StringContent).ReadAsStringAsync();
				}

				if (responseMessage.StatusCode.IsSuccess())
				{
					try
					{
						TResponse response = CommonUtils.DeserializeFromJson<TResponse>(stringResponse);

						return (response == null) ? WebServiceResponse.Create(responseMessage.StatusCode, stringResponse) : WebServiceResponse.Create(responseMessage.StatusCode, response);
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("Failed parsing below string as JSON format (" + ex.Message + ").\r\n" + stringResponse);
						return WebServiceResponse.Create(responseMessage.StatusCode, stringResponse);
					}
				}
				else
				{
					if (string.IsNullOrWhiteSpace(stringResponse))
					{
						throw new WebServiceException(responseMessage);
					}
					else
					{
						throw new WebServiceException(responseMessage, stringResponse);
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods
	}
}
